#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import pytest

from earthkit.data.utils.patterns import Function
from earthkit.data.utils.patterns import Pattern
from earthkit.data.utils.patterns import Variable


@pytest.mark.parametrize(
    "pattern,values,expected_value,error",
    [
        ("level", {"level": "500"}, "500", None),
        ("level:int", {"level": 500}, "500", None),
        ("level:int(%04d)", {"level": 500}, "0500", None),
        ("level:int", {"level": "500"}, None, ValueError),
        ("level:float", {"level": 500}, "500", None),
        ("level:float", {"level": 500.0}, "500", None),
        ("level:float", {"level": 500.1}, "500.1", None),
        ("level:float(%.02f)", {"level": 500.1}, "500.10", None),
        ("level:enum(925,500,700)", {"level": "500"}, "500", None),
        ("level:enum(925,500,700)", {"level": 500}, "500", ValueError),
        ("level:enum(925,500,700)", {"level": "1000"}, None, ValueError),
        ("my_date:date(%Y-%m-%d)", {"my_date": "20200513"}, "2020-05-13", None),
        ("my_date:date(%Y-%m-%d)", {"my_date": datetime.datetime(2020, 5, 13)}, "2020-05-13", None),
        ("my_date:strftime(%Y-%m-%d)", {"my_date": datetime.datetime(2020, 5, 13)}, "2020-05-13", None),
        (
            "my_date:date(%Y%m%d_%H:%M)",
            {"my_date": datetime.datetime(2020, 5, 13, 11, 23, 0)},
            "20200513_11:23",
            None,
        ),
    ],
)
def test_pattern_variable_substitute(pattern, values, expected_value, error):
    v = Variable(pattern)
    if not error:
        assert v.substitute(values) == expected_value
    else:
        with pytest.raises(error):
            v.substitute(values)


@pytest.mark.parametrize(
    "pattern,values,expected_value,error",
    [
        ("level", {"level": "500"}, ["500"], None),
        ("level", {"level": ["500", "300"]}, ["500", "300"], None),
        ("level:int", {"level": 500}, ["500"], None),
        ("level:int", {"level": [500, 300]}, ["500", "300"], None),
        ("level:int", {"level": "500"}, None, ValueError),
        ("level:enum(925,500,700)", {"level": ["500", "700"]}, ["500", "700"], None),
        ("level:enum(925,500,700)", {"level": ["700", "500"]}, ["700", "500"], None),
        ("level:enum(925,500,700)", {"level": 500}, "500", ValueError),
        ("level:enum(925,500,700)", {"level": ["1000", "500"]}, None, ValueError),
        ("my_date:date(%Y-%m-%d)", {"my_date": datetime.datetime(2020, 5, 13)}, ["2020-05-13"], None),
        ("my_date:strftime(%Y-%m-%d)", {"my_date": datetime.datetime(2020, 5, 13)}, ["2020-05-13"], None),
        (
            "my_date:date(%Y%m%d_%H:%M)",
            {"my_date": datetime.datetime(2020, 5, 13, 11, 23, 0)},
            ["20200513_11:23"],
            None,
        ),
    ],
)
def test_pattern_variable_substitute_many(pattern, values, expected_value, error):
    v = Variable(pattern)
    if not error:
        assert v.substitute_many(values) == expected_value
    else:
        with pytest.raises(error):
            v.substitute_many(values)


@pytest.mark.parametrize(
    "pattern,values,expected_value,error",
    [
        ("param|lower", {"param": "TeST"}, "test", None),
    ],
)
def test_pattern_function_substitute(pattern, values, expected_value, error):
    v = Function(pattern)
    if not error:
        assert v.substitute(values) == expected_value
    else:
        with pytest.raises(error):
            v.substitute(values)


@pytest.mark.parametrize(
    "pattern,value,expected_value",
    [
        ("level", "level", {}),
        ("level", "level1", None),
        ("{level}", "500", {"level": "500"}),
        ("_{level}_", "_500_", {"level": "500"}),
        ("_{level}_{date:date(%Y-%m-%d)}.grib", "_500_", None),
        (
            "_{level}_{date:date(%Y-%m-%d)}.grib",
            "_500_2000-01-02.grib",
            {"level": "500", "date": "2000-01-02"},
        ),
    ],
)
def test_pattern_match(pattern, value, expected_value):
    p = Pattern(pattern)
    m = p.match(value)
    if expected_value is None:
        assert m is None
    else:
        assert m.groupdict() == expected_value


@pytest.mark.parametrize(
    "pattern,values,expected_value",
    [
        ("test.{format}", {"format": ["nc", "grib"]}, ["test.nc", "test.grib"]),
        ("test_{id}.grib", {"id": [2, 3, "AA"]}, ["test_2.grib", "test_3.grib", "test_AA.grib"]),
        (
            "test_{my_date:date(%Y-%m-%d)}_{name}.grib",
            {"my_date": datetime.datetime(2020, 5, 13), "name": ["t2", "msl"]},
            ["test_2020-05-13_t2.grib", "test_2020-05-13_msl.grib"],
        ),
        (
            "test_{date:strftimedelta(-6;%Y-%m-%d_%H)}.grib",
            {"date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 11, 6)]},
            ["test_2020-05-10_18.grib", "test_2020-05-11_00.grib"],
        ),
        (
            "test_{date:strftimedelta(60m;%Y-%m-%d_%H)}.grib",
            {"date": [datetime.datetime(2020, 5, 11), datetime.datetime(2020, 5, 11, 6)]},
            ["test_2020-05-11_01.grib", "test_2020-05-11_07.grib"],
        ),
        ("test_{param|lower}", {"param": ["T", "z"]}, ["test_t", "test_z"]),
    ],
)
def test_pattern_core(pattern, values, expected_value):
    p = Pattern(pattern)
    assert p.substitute(values) == expected_value



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from ..readers.zarr import ZarrReader
from . import Source


class ZarrSource(Source):

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self._reader = ZarrReader(self, path, **kwargs)

    def mutate(self):
        source = self._reader.mutate_source()
        if source not in (None, self):
            source._parent = self
            return source

        return self


source = ZarrSource



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import pathlib
import sys

import numpy as np
import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from xr_engine_fixtures import compare_coords  # noqa: E402
from xr_engine_fixtures import load_grib_data  # noqa: E402


@pytest.mark.cache
@pytest.mark.parametrize("engine", ["earthkit", "cfgrib"])
def test_xr_engine_kwargs_unchanged(engine):
    ds = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl_small.grib"))

    _kwargs = {"squeeze": True}
    res = ds.to_xarray(engine=engine, xarray_open_dataset_kwargs=_kwargs)
    assert res is not None
    assert _kwargs == {"squeeze": True}


@pytest.mark.cache
def test_xr_engine_basic():
    ds = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl.grib"))
    res = ds.to_xarray()
    assert res is not None


@pytest.mark.cache
@pytest.mark.parametrize("path_maker", [lambda x: x, lambda x: pathlib.Path(x)])
def test_xr_engine_open_dataset_path(path_maker):

    ds = from_source("sample", "pl.grib")
    path = path_maker(ds.path)

    import xarray as xr

    res = xr.open_dataset(
        path,
        engine="earthkit",
    )
    assert res is not None


@pytest.mark.cache
@pytest.mark.parametrize("api", ["earthkit", "xr"])
def test_xr_engine_detailed_check_1(api):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl.grib"))

    if api == "earthkit":
        ds = ds_ek.to_xarray(
            time_dim_mode="raw",
            decode_times=False,
            decode_timedelta=False,
            add_valid_time_coord=False,
            dim_name_from_role_name=False,
        )
    else:
        import xarray as xr

        ds = xr.open_dataset(
            ds_ek.path,
            engine="earthkit",
            time_dim_mode="raw",
            decode_times=False,
            decode_timedelta=False,
            add_valid_time_coord=False,
            dim_name_from_role_name=False,
        )

    assert ds is not None

    # dataset
    lats = np.linspace(90, -90, 19)
    lons = np.linspace(0, 350, 36)
    data_vars = ["r", "t", "u", "v", "z"]

    coords_ref_full = {
        "date": np.array([20240603, 20240604]),
        "time": np.array([0, 1200]),
        "step_timedelta": [0, 6],
        "levelist": np.array([300, 400, 500, 700, 850, 1000]),
        "latitude": lats,
        "longitude": lons,
    }

    dims_ref_full = {
        "date": 2,
        "time": 2,
        "step_timedelta": 2,
        "levelist": 6,
        "latitude": 19,
        "longitude": 36,
    }

    assert len(ds.dims) == len(dims_ref_full)
    compare_coords(ds, coords_ref_full)
    assert [v for v in ds.data_vars] == data_vars

    # data variable
    assert ds["u"].shape == (2, 2, 2, 6, 19, 36)
    assert ds["u"].values.shape == (2, 2, 2, 6, 19, 36)
    assert ds["u"].as_numpy().shape == (2, 2, 2, 6, 19, 36)
    assert ds["u"].to_numpy().shape == (2, 2, 2, 6, 19, 36)
    r = ds["u"]
    compare_coords(r, coords_ref_full)

    # sel() on dataset
    r = ds.sel(date=20240603, time=[0, 1200])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    compare_coords(r, coords_ref)
    assert [v for v in r.data_vars] == data_vars

    # sel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 19, 36)
    r1 = r["u"].sel(step_timedelta=6, levelist=[1000, 300])
    assert r1.shape == (2, 2, 19, 36)
    coords_ref["step_timedelta"] = [6]
    coords_ref["levelist"] = np.array([1000, 300])
    compare_coords(r1, coords_ref)

    # isel() on dataset
    r = ds.isel(date=0, time=[0, 1])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    compare_coords(r, coords_ref)
    assert [v for v in r.data_vars] == data_vars

    # isel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 19, 36)
    r1 = r["u"].isel(step_timedelta=1, levelist=[0, -1])
    assert r1.shape == (2, 2, 19, 36)
    coords_ref["step_timedelta"] = [6]
    coords_ref["levelist"] = np.array([300, 1000])
    compare_coords(r1, coords_ref)

    # slicing of data variable
    da = ds["u"]

    r = da[:, 0]
    assert r.shape == (2, 2, 6, 19, 36)
    assert r.values.shape == (2, 2, 6, 19, 36)
    assert r.to_numpy().shape == (2, 2, 6, 19, 36)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    compare_coords(r, coords_ref)

    r = da[:, 0, :, 3:5]
    assert r.shape == (2, 2, 2, 19, 36)
    assert r.values.shape == (2, 2, 2, 19, 36)
    assert r.to_numpy().shape == (2, 2, 2, 19, 36)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["levelist"] = np.array([700, 850])
    compare_coords(r, coords_ref)

    r = da.loc[:, 0, :, [700, 850]]
    assert r.shape == (2, 2, 2, 19, 36)
    assert r.values.shape == (2, 2, 2, 19, 36)
    assert r.to_numpy().shape == (2, 2, 2, 19, 36)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["levelist"] = np.array([700, 850])
    compare_coords(r, coords_ref)

    # lat-lon
    da = ds["t"]

    r = da[:, 0, :, 2, 9, 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)

    r = da[:, 0, :, 2, 9:12, :2]
    assert r.shape == (2, 2, 3, 2)
    vals_ref = np.array(
        [
            [
                [
                    [269.00918579, 269.31680298],
                    [269.70254517, 269.81387329],
                    [267.50527954, 266.83828735],
                ],
                [
                    [268.78610229, 268.80758667],
                    [269.52731323, 269.75680542],
                    [266.61813354, 267.12106323],
                ],
            ],
            [
                [
                    [268.57771301, 269.03767395],
                    [269.33357239, 269.56111145],
                    [264.75154114, 266.55036926],
                ],
                [
                    [268.08932495, 268.35983276],
                    [269.01803589, 269.02389526],
                    [264.29733276, 266.08248901],
                ],
            ],
        ]
    )
    assert np.allclose(r.values, vals_ref)

    r = da.loc[:, 0, :, 500, 0, 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)


@pytest.mark.cache
@pytest.mark.parametrize("api", ["earthkit", "xr"])
def test_xr_engine_detailed_check_2(api):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl.grib"))

    if api == "earthkit":
        ds = ds_ek.to_xarray(
            time_dim_mode="raw",
            decode_times=False,
            decode_timedelta=False,
            add_valid_time_coord=False,
            dim_name_from_role_name=True,
        )
    else:
        import xarray as xr

        ds = xr.open_dataset(
            ds_ek.path,
            engine="earthkit",
            time_dim_mode="raw",
            decode_times=False,
            decode_timedelta=False,
            add_valid_time_coord=False,
            dim_name_from_role_name=True,
        )

    assert ds is not None

    # dataset
    lats = np.linspace(90, -90, 19)
    lons = np.linspace(0, 350, 36)
    data_vars = ["r", "t", "u", "v", "z"]

    coords_ref_full = {
        "date": np.array([20240603, 20240604]),
        "time": np.array([0, 1200]),
        "step": [0, 6],
        "level": np.array([300, 400, 500, 700, 850, 1000]),
        "latitude": lats,
        "longitude": lons,
    }

    dims_ref_full = {
        "date": 2,
        "time": 2,
        "step": 2,
        "level": 6,
        "latitude": 19,
        "longitude": 36,
    }

    assert len(ds.dims) == len(dims_ref_full)
    compare_coords(ds, coords_ref_full)
    assert [v for v in ds.data_vars] == data_vars

    # data variable
    assert ds["u"].shape == (2, 2, 2, 6, 19, 36)
    assert ds["u"].values.shape == (2, 2, 2, 6, 19, 36)
    assert ds["u"].as_numpy().shape == (2, 2, 2, 6, 19, 36)
    assert ds["u"].to_numpy().shape == (2, 2, 2, 6, 19, 36)
    r = ds["u"]
    compare_coords(r, coords_ref_full)

    # sel() on dataset
    r = ds.sel(date=20240603, time=[0, 1200])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    compare_coords(r, coords_ref)
    assert [v for v in r.data_vars] == data_vars

    # sel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 19, 36)
    r1 = r["u"].sel(step=6, level=[1000, 300])
    assert r1.shape == (2, 2, 19, 36)
    coords_ref["step"] = [6]
    coords_ref["level"] = np.array([1000, 300])
    compare_coords(r1, coords_ref)

    # isel() on dataset
    r = ds.isel(date=0, time=[0, 1])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    compare_coords(r, coords_ref)
    assert [v for v in r.data_vars] == data_vars

    # isel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 19, 36)
    r1 = r["u"].isel(step=1, level=[0, -1])
    assert r1.shape == (2, 2, 19, 36)
    coords_ref["step"] = [6]
    coords_ref["level"] = np.array([300, 1000])
    compare_coords(r1, coords_ref)

    # slicing of data variable
    da = ds["u"]

    r = da[:, 0]
    assert r.shape == (2, 2, 6, 19, 36)
    assert r.values.shape == (2, 2, 6, 19, 36)
    assert r.to_numpy().shape == (2, 2, 6, 19, 36)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    compare_coords(r, coords_ref)

    r = da[:, 0, :, 3:5]
    assert r.shape == (2, 2, 2, 19, 36)
    assert r.values.shape == (2, 2, 2, 19, 36)
    assert r.to_numpy().shape == (2, 2, 2, 19, 36)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["level"] = np.array([700, 850])
    compare_coords(r, coords_ref)

    r = da.loc[:, 0, :, [700, 850]]
    assert r.shape == (2, 2, 2, 19, 36)
    assert r.values.shape == (2, 2, 2, 19, 36)
    assert r.to_numpy().shape == (2, 2, 2, 19, 36)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["level"] = np.array([700, 850])
    compare_coords(r, coords_ref)

    # lat-lon
    da = ds["t"]

    r = da[:, 0, :, 2, 9, 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)

    r = da[:, 0, :, 2, 9:12, :2]
    assert r.shape == (2, 2, 3, 2)
    vals_ref = np.array(
        [
            [
                [
                    [269.00918579, 269.31680298],
                    [269.70254517, 269.81387329],
                    [267.50527954, 266.83828735],
                ],
                [
                    [268.78610229, 268.80758667],
                    [269.52731323, 269.75680542],
                    [266.61813354, 267.12106323],
                ],
            ],
            [
                [
                    [268.57771301, 269.03767395],
                    [269.33357239, 269.56111145],
                    [264.75154114, 266.55036926],
                ],
                [
                    [268.08932495, 268.35983276],
                    [269.01803589, 269.02389526],
                    [264.29733276, 266.08248901],
                ],
            ],
        ]
    )
    assert np.allclose(r.values, vals_ref)

    r = da.loc[:, 0, :, 500, 0, 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)


@pytest.mark.cache
@pytest.mark.parametrize("stream", [False, True])
@pytest.mark.parametrize("lazy_load", [False, True])
@pytest.mark.parametrize("release_source", [False, True])
@pytest.mark.parametrize("direct_backend", [False, True])
def test_xr_engine_detailed_flatten_check_1(stream, lazy_load, release_source, direct_backend):
    filename = "test-data/xr_engine/level/pl.grib"
    ds_ek, ds_ek_ref = load_grib_data(filename, "url", stream=stream)

    kwargs = {
        "xarray_open_dataset_kwargs": {
            "backend_kwargs": {
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "flatten_values": True,
                "add_valid_time_coord": False,
                "lazy_load": lazy_load,
                "release_source": release_source,
                "direct_backend": direct_backend,
                "dim_name_from_role_name": False,
            }
        }
    }

    ds = ds_ek.to_xarray(**kwargs)
    assert ds is not None

    # dataset
    ll = ds_ek_ref[0].to_latlon(flatten=True)
    lats = ll["lat"]
    lons = ll["lon"]
    data_vars = ["r", "t", "u", "v", "z"]

    coords_ref_full = {
        "date": np.array([20240603, 20240604]),
        "time": np.array([0, 1200]),
        "step_timedelta": np.array([0, 6]),
        "levelist": np.array([300, 400, 500, 700, 850, 1000]),
        "latitude": lats,
        "longitude": lons,
    }

    dims_ref_full = {
        "date": 2,
        "time": 2,
        "step_timedelta": 2,
        "levelist": 6,
        "values": 684,
    }

    assert len(ds.dims) == len(dims_ref_full)
    assert len(ds.coords) == len(coords_ref_full)
    for k, v in coords_ref_full.items():
        assert np.allclose(ds.coords[k].values, v)
    assert [v for v in ds.data_vars] == data_vars

    # data variable
    assert ds["u"].shape == (2, 2, 2, 6, 684)
    assert ds["u"].values.shape == (2, 2, 2, 6, 684)
    assert ds["u"].as_numpy().shape == (2, 2, 2, 6, 684)
    assert ds["u"].to_numpy().shape == (2, 2, 2, 6, 684)
    r = ds["u"]
    assert len(r.coords) == len(coords_ref_full)
    for k, v in coords_ref_full.items():
        assert np.allclose(r.coords[k].values, v)

    # sel() on dataset
    r = ds.sel(date=20240603, time=[0, 1200])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    assert len(r.coords) == len(coords_ref)
    for k, v in coords_ref.items():
        assert np.allclose(r.coords[k].values, v)
    assert [v for v in r.data_vars] == data_vars

    # sel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 684)
    r1 = r["u"].sel(step_timedelta=6, levelist=[1000, 300])
    assert r1.shape == (2, 2, 684)
    coords_ref["step_timedelta"] = np.array([6])
    coords_ref["levelist"] = np.array([1000, 300])
    assert len(r1.coords) == len(coords_ref)
    for k, v in coords_ref.items():
        assert np.allclose(r1.coords[k].values, v)

    # isel() on dataset
    r = ds.isel(date=0, time=[0, 1])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    assert len(r.coords) == len(coords_ref)
    for k, v in coords_ref.items():
        assert np.allclose(r.coords[k].values, v)
    assert [v for v in r.data_vars] == data_vars

    # isel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 684)
    r1 = r["u"].isel(step_timedelta=1, levelist=[0, -1])
    assert r1.shape == (2, 2, 684)
    coords_ref["step_timedelta"] = np.array([6])
    coords_ref["levelist"] = np.array([300, 1000])
    assert len(r1.coords) == len(coords_ref)
    for k, v in coords_ref.items():
        assert np.allclose(r1.coords[k].values, v)

    # slicing of data variable
    da = ds["u"]

    r = da[:, 0]
    assert r.shape == (2, 2, 6, 684)
    assert r.values.shape == (2, 2, 6, 684)
    assert r.to_numpy().shape == (2, 2, 6, 684)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    for k, v in coords_ref.items():
        assert np.allclose(r.coords[k].values, v)

    r = da[:, 0, :, 3:5]
    assert r.shape == (2, 2, 2, 684)
    assert r.values.shape == (2, 2, 2, 684)
    assert r.to_numpy().shape == (2, 2, 2, 684)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["levelist"] = np.array([700, 850])
    for k, v in coords_ref.items():
        assert np.allclose(r.coords[k].values, v)

    r = da.loc[:, 0, :, [700, 850]]
    assert r.shape == (2, 2, 2, 684)
    assert r.values.shape == (2, 2, 2, 684)
    assert r.to_numpy().shape == (2, 2, 2, 684)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["levelist"] = np.array([700, 850])
    for k, v in coords_ref.items():
        assert np.allclose(r.coords[k].values, v)

    # level=500, lat=0, lon=0
    da = ds["t"]

    r = da[:, 0, :, 2, 9 * 36 + 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)

    r = da[:, 0, :, 2, [9 * 36, 10 * 36, 11 * 36]]
    assert r.shape == (2, 2, 3)
    vals_ref = np.array(
        [
            [
                [269.00918579, 269.70254517, 267.50527954],
                [268.78610229, 269.52731323, 266.61813354],
            ],
            [
                [268.57771301, 269.33357239, 264.75154114],
                [268.08932495, 269.01803589, 264.29733276],
            ],
        ]
    )

    v_ek = ds_ek_ref.sel(param="t", time=0, levelist=500).to_numpy(flatten=True)
    assert np.allclose(r.values.flatten(), v_ek[:, [9 * 36, 10 * 36, 11 * 36]].flatten())
    assert np.allclose(r.values, vals_ref)

    r = da.loc[:, 0, :, 500, 9 * 36 + 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)


@pytest.mark.cache
@pytest.mark.parametrize("stream", [False, True])
@pytest.mark.parametrize("lazy_load", [False, True])
@pytest.mark.parametrize("release_source", [False, True])
@pytest.mark.parametrize("direct_backend", [False, True])
def test_xr_engine_detailed_flatten_check_2(stream, lazy_load, release_source, direct_backend):
    filename = "test-data/xr_engine/level/pl.grib"
    ds_ek, ds_ek_ref = load_grib_data(filename, "url", stream=stream)

    kwargs = {
        "xarray_open_dataset_kwargs": {
            "backend_kwargs": {
                "time_dim_mode": "raw",
                "decode_times": False,
                "decode_timedelta": False,
                "flatten_values": True,
                "add_valid_time_coord": False,
                "lazy_load": lazy_load,
                "release_source": release_source,
                "direct_backend": direct_backend,
                "dim_name_from_role_name": True,
            }
        }
    }

    ds = ds_ek.to_xarray(**kwargs)
    assert ds is not None

    # dataset
    ll = ds_ek_ref[0].to_latlon(flatten=True)
    lats = ll["lat"]
    lons = ll["lon"]
    data_vars = ["r", "t", "u", "v", "z"]

    coords_ref_full = {
        "date": np.array([20240603, 20240604]),
        "time": np.array([0, 1200]),
        "step": np.array([0, 6]),
        "level": np.array([300, 400, 500, 700, 850, 1000]),
        "latitude": lats,
        "longitude": lons,
    }

    dims_ref_full = {
        "date": 2,
        "time": 2,
        "step": 2,
        "level": 6,
        "values": 684,
    }

    assert len(ds.dims) == len(dims_ref_full)
    compare_coords(ds, coords_ref_full)
    assert [v for v in ds.data_vars] == data_vars

    # data variable
    assert ds["u"].shape == (2, 2, 2, 6, 684)
    assert ds["u"].values.shape == (2, 2, 2, 6, 684)
    assert ds["u"].as_numpy().shape == (2, 2, 2, 6, 684)
    assert ds["u"].to_numpy().shape == (2, 2, 2, 6, 684)
    r = ds["u"]
    compare_coords(r, coords_ref_full)

    # sel() on dataset
    r = ds.sel(date=20240603, time=[0, 1200])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    compare_coords(r, coords_ref)
    assert [v for v in r.data_vars] == data_vars

    # sel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 684)
    r1 = r["u"].sel(step=6, level=[1000, 300])
    assert r1.shape == (2, 2, 684)
    coords_ref["step"] = np.array([6])
    coords_ref["level"] = np.array([1000, 300])
    compare_coords(r1, coords_ref)

    # isel() on dataset
    r = ds.isel(date=0, time=[0, 1])
    coords_ref = dict(coords_ref_full)
    coords_ref["date"] = np.array([20240603])
    compare_coords(r, coords_ref)
    assert [v for v in r.data_vars] == data_vars

    # isel() on data variable of filtered dataset
    assert r["u"].shape == (2, 2, 6, 684)
    r1 = r["u"].isel(step=1, level=[0, -1])
    assert r1.shape == (2, 2, 684)
    coords_ref["step"] = np.array([6])
    coords_ref["level"] = np.array([300, 1000])
    compare_coords(r1, coords_ref)

    # slicing of data variable
    da = ds["u"]

    r = da[:, 0]
    assert r.shape == (2, 2, 6, 684)
    assert r.values.shape == (2, 2, 6, 684)
    assert r.to_numpy().shape == (2, 2, 6, 684)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    compare_coords(r, coords_ref)

    r = da[:, 0, :, 3:5]
    assert r.shape == (2, 2, 2, 684)
    assert r.values.shape == (2, 2, 2, 684)
    assert r.to_numpy().shape == (2, 2, 2, 684)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["level"] = np.array([700, 850])
    compare_coords(r, coords_ref)

    r = da.loc[:, 0, :, [700, 850]]
    assert r.shape == (2, 2, 2, 684)
    assert r.values.shape == (2, 2, 2, 684)
    assert r.to_numpy().shape == (2, 2, 2, 684)
    dims_ref = dict(dims_ref_full)
    dims_ref.pop("time")
    assert len(r.dims) == len(dims_ref)
    coords_ref = dict(coords_ref_full)
    coords_ref["time"] = np.array([0])
    coords_ref["level"] = np.array([700, 850])
    compare_coords(r, coords_ref)

    # level=500, lat=0, lon=0
    da = ds["t"]

    r = da[:, 0, :, 2, 9 * 36 + 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)

    r = da[:, 0, :, 2, [9 * 36, 10 * 36, 11 * 36]]
    assert r.shape == (2, 2, 3)
    vals_ref = np.array(
        [
            [
                [269.00918579, 269.70254517, 267.50527954],
                [268.78610229, 269.52731323, 266.61813354],
            ],
            [
                [268.57771301, 269.33357239, 264.75154114],
                [268.08932495, 269.01803589, 264.29733276],
            ],
        ]
    )

    v_ek = ds_ek_ref.sel(param="t", time=0, levelist=500).to_numpy(flatten=True)
    assert np.allclose(r.values.flatten(), v_ek[:, [9 * 36, 10 * 36, 11 * 36]].flatten())
    assert np.allclose(r.values, vals_ref)

    r = da.loc[:, 0, :, 500, 9 * 36 + 0]
    assert r.shape == (2, 2)
    vals_ref = np.array([[269.00918579, 268.78610229], [268.57771301, 268.08932495]])
    assert np.allclose(r.values, vals_ref)


@pytest.mark.cache
@pytest.mark.parametrize(
    "kwargs",
    [
        {"split_dims": ["step"]},
        {"split_dims": None},
        {"direct_backend": None},
        {"direct_backend": True},
        {"direct_backend": False},
    ],
)
def test_xr_engine_invalid_kwargs(kwargs):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "level", "pl.grib"))

    import xarray as xr

    with pytest.raises(TypeError):
        xr.open_dataset(
            ds_ek.path,
            engine="earthkit",
            time_dim_mode="raw",
            **kwargs,
        )


@pytest.mark.cache
@pytest.mark.parametrize(
    "dtype,expected_dtype",
    [
        (np.float32, np.float32),
        ("float32", np.float32),
        (np.float64, np.float64),
        ("float64", np.float64),
    ],
)
def test_xr_engine_dtype(dtype, expected_dtype):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))

    ds = ds_ek.to_xarray(dtype=dtype)
    assert ds["t"].data.dtype == expected_dtype
    assert ds["t"].values.dtype == expected_dtype


@pytest.mark.cache
def test_xr_engine_single_field():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek[0]

    ds = ds_ek.to_xarray(
        time_dim_mode="raw",
        decode_times=False,
        decode_timedelta=False,
        add_valid_time_coord=False,
        squeeze=True,
    )

    vals_ref = ds_ek.to_numpy()

    lats = np.linspace(90, -90, 19)
    lons = np.linspace(0, 350, 36)

    var_attrs_ref = {
        "standard_name": "air_temperature",
        "long_name": "Temperature",
        "units": "K",
    }

    global_attrs_ref = {
        "param": "t",
        "paramId": 130,
        "class": "od",
        "stream": "oper",
        "levtype": "pl",
        "type": "fc",
        "expver": "0001",
        "date": 20240603,
        "time": 0,
        "domain": "g",
        "number": 0,
        "levelist": 1000,
        "Conventions": "CF-1.8",
        "institution": "ECMWF",
    }

    assert ds.attrs == global_attrs_ref

    data_vars = ["t"]

    coords_ref_full = {
        "latitude": lats,
        "longitude": lons,
    }

    dims_ref_full = {
        "latitude": 19,
        "longitude": 36,
    }

    assert len(ds.dims) == len(dims_ref_full)
    assert len(ds.coords) == len(coords_ref_full)
    for k, v in coords_ref_full.items():
        assert np.allclose(ds.coords[k].values, v)
    assert [v for v in ds.data_vars] == data_vars

    da = ds["t"]

    for k, v in var_attrs_ref.items():
        assert da.attrs[k] == v

    r = da[:, :]
    r.shape == (19, 36)
    assert np.allclose(r.values, vals_ref)

    r = da[2:4, 6:8]
    assert r.shape == (2, 2)
    assert np.allclose(r.values, vals_ref[2:4, 6:8])

    v = da.sel(latitude=0, longitude=0, method="nearest")
    assert np.allclose(v, vals_ref[9, 0])


@pytest.mark.cache
@pytest.mark.parametrize("add", [False, True])
def test_xr_engine_add_earthkit_attrs_1(add):
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek[0]

    ds = ds_ek.to_xarray(
        time_dim_mode="raw",
        decode_times=False,
        decode_timedelta=False,
        add_valid_time_coord=False,
        add_earthkit_attrs=add,
    )

    if add:
        assert "_earthkit" in ds["t"].attrs
    else:
        assert "_earthkit" not in ds["t"].attrs


@pytest.mark.cache
def test_xr_engine_add_earthkit_attrs_2():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))
    ds_ek = ds_ek[0]

    ds = ds_ek.to_xarray(
        add_earthkit_attrs=False,
    )

    assert ds
    assert "_earthkit" not in ds["t"].attrs



import earthkit.data as ekd

ds = ekd.from_source("sample", "test.grib")

# case 1
o = ekd.new_grib_output("res.grib", shortName="2t")
for field in ds:
    o.write(None, template=field)
o.close()

# case 2
o = ekd.new_grib_output("res.grib")
for field in ds:
    o.write(None, shortName="2t", template=field)
o.close()

# case 3
with ekd.new_grib_output("res.grib") as o:
    for field in ds:
        o.write(None, shortName="2t", template=field)

# case 4
with ekd.new_grib_output("res.grib") as o:
    for field in ds:
        values = field.values + 1
        o.write(values, shortName="2t", template=field)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime

import numpy as np
import pytest


def _build_list(prototype):
    return [
        {"param": "t", "levelist": 500, **prototype},
        {"param": "t", "levelist": 850, **prototype},
        {"param": "u", "levelist": 500, **prototype},
        {"param": "u", "levelist": 850, **prototype},
        {"param": "d", "levelist": 850, **prototype},
        {"param": "d", "levelist": 600, **prototype},
    ]


@pytest.fixture
def lod_input_data(request):
    return request.param


@pytest.fixture
def lod_distinct_ll_list_values():
    prototype = {
        "latitudes": [-10.0, 0.0, 10.0],
        "longitudes": [20, 40.0],
        "values": [1, 2, 3, 4, 5, 6],
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_distinct_ll():
    prototype = {
        "latitudes": np.array([-10.0, 0.0, 10.0]),
        "longitudes": np.array([20, 40.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_flat():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_flat_invalid():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([1, 2, 3, 4, 5]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_flat_10x10():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 30.0, 20.0, 30.0, 20.0, 30.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_2D_all():
    prototype = {
        "latitudes": np.array([[-10.0, -10.0], [0.0, 0.0], [10.0, 10.0]]),
        "longitudes": np.array([[20.0, 40.0], [20.0, 40.0], [20.0, 40.0]]),
        "values": np.array([[1, 2], [3, 4], [5, 6]]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_2D_ll():
    prototype = {
        "latitudes": np.array([[-10.0, -10.0], [0.0, 0.0], [10.0, 10.0]]),
        "longitudes": np.array([[20.0, 40.0], [20.0, 40.0], [20.0, 40.0]]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_2D_values():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([[1, 2], [3, 4], [5, 6]]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_forecast_1():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "base_datetime": "2018-08-01T03:00:00",
        "step": 6,
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_forecast_2():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "valid_datetime": "2018-08-01T09:00:00",
        "step": 6,
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_forecast_3():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "forecast_reference_time": "2018-08-01T03:00:00",
        "step": datetime.timedelta(hours=6),
    }
    return _build_list(prototype)


@pytest.fixture
def lod_ll_forecast_4():
    prototype = {
        "latitudes": np.array([-10.0, -10.0, 0.0, 0.0, 10.0, 10.0]),
        "longitudes": np.array([20.0, 40.0, 20.0, 40.0, 20.0, 40.0]),
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "date": 20180801,
        "time": 300,
        "step": datetime.timedelta(hours=6),
    }
    return _build_list(prototype)


@pytest.fixture
def lod_no_latlon():
    prototype = {
        "values": np.array([1, 2, 3, 4, 5, 6]),
        "valid_datetime": "2018-08-01T09:00:00",
    }
    return _build_list(prototype)


def build_lod_fieldlist(lod, mode):
    from earthkit.data import from_source
    from earthkit.data.indexing.fieldlist import SimpleFieldList
    from earthkit.data.sources.array_list import ArrayField

    if mode == "list-of-dicts":
        return from_source("list-of-dicts", lod)
    elif mode == "loop":
        ds = SimpleFieldList()
        for f in lod:
            ds.append(ArrayField(f["values"], f))
        return ds



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pytest

from earthkit.data import from_source
from earthkit.data.testing import earthkit_examples_file


@pytest.mark.parametrize(
    "parts,expected_meta",
    [
        ((0, 1596), ["02836"]),  # message length
        # ([(0, 1596)], ["02836"]),  # message length
        # ([(0, 1601)], ["02836"]),  # message length + extra bytes
        # ([(2904, 1166)], ["01415"]),
        # ([(0, 1596)], ["02836"]),  # shorter than message
        # ([(2904, 2442)], ["01415", "01001"]),
        # ([(2904, 1166), (5346, 1216)], ["01415", "01152"]),
        # ([(2904, 1166), (5346, 121)], ["01415"]),  # second part shorter than message
    ],
)
def test_bufr_single_file_parts(parts, expected_meta):
    ds = from_source("file", earthkit_examples_file("temp_10.bufr"), parts=parts)

    assert len(ds) == len(expected_meta)
    if len(ds) > 0:
        assert ds.metadata(("ident")) == expected_meta


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest

from earthkit.data import cache
from earthkit.data import from_source
from earthkit.data import settings
from earthkit.data.core.caching import cache_file
from earthkit.data.core.temporary import temp_directory
from earthkit.data.testing import earthkit_examples_file


def check_cache_files(dir_path, managed=True):
    def touch(target, args):
        assert args["foo"] in (1, 2)
        with open(target, "w"):
            pass

    path1 = cache_file(
        "test_cache",
        touch,
        {"foo": 1},
        extension=".test",
    )

    path2 = cache_file(
        "test_cache",
        touch,
        {"foo": 2},
        extension=".test",
    )

    assert os.path.exists(path1)
    assert os.path.exists(path2)
    assert os.path.dirname(path1) == dir_path
    assert os.path.dirname(path1) == dir_path
    assert path1 != path2

    if managed:
        cnt = 0
        for f in cache.entries():
            if f["owner"] == "test_cache":
                cnt += 1

        assert cnt == 2


@pytest.mark.cache
def test_cache_1_setting():
    with settings.temporary():
        settings.set("maximum-cache-disk-usage", "99%")
        cache.purge(matcher=lambda e: ["owner"] == "test_cache")
        check_cache_files(settings.get("user-cache-directory"))


# 1GB ram disk on MacOS (blocks of 512 bytes)
# diskutil erasevolume HFS+ "RAMDisk" `hdiutil attach -nomount ram://2097152`
@pytest.mark.skipif(not os.path.exists("/Volumes/RAMDisk"), reason="No RAM disk")
def test_cache_4_setting():
    with settings.temporary():
        settings.set("cache-directory", "/Volumes/RAMDisk/earthkit_data")
        settings.set("maximum-cache-disk-usage", "90%")
        for n in range(10):
            from_source("dummy-source", "zeros", size=100 * 1024 * 1024, n=n)


def test_cache_policy_setting():
    with temp_directory() as user_dir:
        # cache = user dir
        with settings.temporary():
            settings.set({"cache-policy": "user", "user-cache-directory": user_dir})
            assert settings.get("cache-policy") == "user"
            assert settings.get("user-cache-directory") == user_dir
            assert cache.policy.managed() is True
            cache_dir = cache.policy.directory()
            assert cache_dir == user_dir
            assert os.path.exists(cache_dir)
            check_cache_files(cache_dir)

            # cache = temporary with auto generated path
            with settings.temporary({"cache-policy": "temporary", "temporary-cache-directory-root": None}):
                assert settings.get("cache-policy") == "temporary"
                assert settings.get("temporary-cache-directory-root") is None
                assert cache.policy.managed() is True
                cache_dir = cache.policy.directory()
                assert os.path.exists(cache_dir)
                check_cache_files(cache_dir)

            # cache = user dir (again)
            assert settings.get("cache-policy") == "user"
            assert settings.get("user-cache-directory") == user_dir
            assert cache.policy.managed() is True
            cache_dir = cache.policy.directory()
            assert cache_dir == user_dir
            assert os.path.exists(cache_dir)
            check_cache_files(cache_dir)

            # cache = temporary with user defined root path
            with temp_directory() as root_dir:
                with settings.temporary(
                    {
                        "cache-policy": "temporary",
                        "temporary-cache-directory-root": root_dir,
                    }
                ):
                    assert settings.get("cache-policy") == "temporary"
                    assert settings.get("temporary-cache-directory-root") == root_dir
                    assert cache.policy.managed() is True
                    cache_dir = cache.policy.directory()
                    assert os.path.exists(cache_dir)
                    os.path.dirname(cache_dir) == root_dir
                    check_cache_files(cache_dir)

            # cache = off
            with settings.temporary("cache-policy", "off"):
                assert settings.get("cache-policy") == "off"
                assert settings.get("user-cache-directory") == user_dir
                assert cache.policy.managed() is False

                cache_dir = cache.policy.directory()
                assert os.path.exists(cache_dir)
                check_cache_files(cache_dir, managed=False)

            # cache = user dir (again)
            assert settings.get("cache-policy") == "user"
            assert settings.get("user-cache-directory") == user_dir
            assert cache.policy.managed() is True
            cache_dir = cache.policy.directory()
            assert cache_dir == user_dir
            assert os.path.exists(cache_dir)
            check_cache_files(cache_dir)


def test_url_source_no_cache_setting():
    with settings.temporary("cache-policy", "off"):
        ds = from_source(
            "url",
            "https://get.ecmwf.int/repository/test-data/earthkit-data/examples/test.grib",
        )
        assert len(ds) == 2


def test_grib_no_cache_setting():
    with settings.temporary("cache-policy", "off"):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        f = ds[3]
        assert f.metadata("param") == "t"


@pytest.mark.parametrize("index_cache", [True, False])
def test_grib_offset_index_cache_setting(index_cache):
    s = {"cache-policy": "temporary", "use-message-position-index-cache": index_cache}
    with settings.temporary(s):
        ds = from_source("file", earthkit_examples_file("tuv_pl.grib"))
        assert len(ds) == 18

        f = ds[3]
        assert f.metadata("param") == "t", f"index-cache={index_cache}"


# See github #155. This test can hang so we must set a timeout.
@pytest.mark.no_cache_init
@pytest.mark.timeout(20)
def test_cache_with_log_debug_setting(caplog):
    import logging

    # the cache must not be initialised at this point
    assert cache._policy is None
    assert cache._manager is None

    caplog.set_level(logging.DEBUG)
    LOG = logging.getLogger(__name__)

    class A:
        def __repr__(self):
            d = cache.directory()
            return d

    a = A()
    LOG.debug(f"dir {a}")
    # NOTE: if we use "%s" formatting e.g. "LOG.debug("dir %s", a)"
    # the problem still occurs!


@pytest.mark.cache
def test_cache_zip_file_overwritten_1_setting():
    with temp_directory() as tmp_dir:
        import shutil
        import zipfile

        # copy input data to work dir
        grb1_path = os.path.join(tmp_dir, "test.grib")
        shutil.copyfile(earthkit_examples_file("test.grib"), grb1_path)

        grb2_path = os.path.join(tmp_dir, "test6.grib")
        shutil.copyfile(earthkit_examples_file("test6.grib"), grb2_path)

        # first pass
        zip_path = os.path.join(tmp_dir, "test.zip")
        with zipfile.ZipFile(zip_path, "w") as zip_object:
            zip_object.write(grb1_path)

        ds = from_source("file", zip_path)
        assert len(ds) == 2
        ds_path = ds.path

        # second pass - same zip file, the grib should be read
        #  from the cache
        ds1 = from_source("file", zip_path)
        assert len(ds1) == 2
        assert ds1.path == ds_path

        # third pass - same zipfile path with different contents
        with zipfile.ZipFile(zip_path, "w") as zip_object:
            zip_object.write(grb2_path)

        ds2 = from_source("file", zip_path)
        assert len(ds2) == 6
        assert ds2.path != ds_path


@pytest.mark.cache
def test_cache_zip_file_changed_modtime_setting():
    with temp_directory() as tmp_dir:
        import shutil
        import zipfile

        # copy input data to work dir
        grb1_path = os.path.join(tmp_dir, "test.grib")
        shutil.copyfile(earthkit_examples_file("test.grib"), grb1_path)

        # first pass
        zip_path = os.path.join(tmp_dir, "test.zip")
        with zipfile.ZipFile(zip_path, "w") as zip_object:
            zip_object.write(grb1_path)

        ds = from_source("file", zip_path)
        assert len(ds) == 2
        ds_path = ds.path

        # second pass - changed modtime
        st = os.stat(zip_path)
        m_time = (st.st_atime_ns + 10, st.st_mtime_ns + 10)
        os.utime(zip_path, ns=m_time)
        ds2 = from_source("file", zip_path)
        assert len(ds2) == 2
        assert ds2.path != ds_path


@pytest.mark.parametrize("policy", ["user", "temporary"])
def test_cache_management_setting(policy):
    with temp_directory() as tmp_dir_path:
        with settings.temporary():
            if policy == "user":
                settings.set({"cache-policy": "user", "user-cache-directory": tmp_dir_path})
                assert cache.directory() == tmp_dir_path
            elif policy == "temporary":
                settings.set(
                    {
                        "cache-policy": "temporary",
                        "temporary-cache-directory-root": tmp_dir_path,
                    }
                )
                assert os.path.dirname(cache.directory()) == tmp_dir_path
            else:
                assert False

            data_size = 10 * 1024

            # create 3 files existing only in the cache
            r = []
            for n in range(3):
                r.append(from_source("dummy-source", "zeros", size=data_size, n=n))

            for ds in r:
                assert os.path.exists(ds.path)
                assert os.path.dirname(ds.path) == cache.directory()

            # check cache contents
            num, size = cache.summary_dump_database()
            assert num == 3
            assert size == 3 * data_size
            assert len(cache.entries()) == 3

            for i, x in enumerate(cache.entries()):
                assert x["size"] == data_size
                assert x["owner"] == "dummy-source"
                assert x["args"] == {"size": data_size, "n": i}
                latest_path = x["path"]

            # limit cache size so that only one file should remain
            settings.set({"maximum-cache-size": "12K", "maximum-cache-disk-usage": None})

            num, size = cache.summary_dump_database()
            assert num == 1
            assert size == data_size
            assert len(cache.entries()) == 1
            for x in cache.entries():
                assert x["size"] == data_size
                assert x["owner"] == "dummy-source"
                assert x["args"] == {"size": data_size, "n": 2}
                x["path"] == latest_path
                break

            # purge the cache
            r = None
            cache.purge()
            num, size = cache.summary_dump_database()
            assert num == 0
            assert size == 0
            assert len(cache.entries()) == 0


@pytest.mark.cache
def test_cache_force_setting():
    import time

    def _force_true(args, path, owner_data):
        time.sleep(0.001)
        return True

    def _force_false(args, path, owner_data):
        time.sleep(0.001)
        return False

    data_size = 10 * 1024
    ds = from_source("dummy-source", "zeros", size=data_size, n=0)
    st = os.stat(ds.path)
    m_time_ref = st.st_mtime_ns

    ds1 = from_source("dummy-source", "zeros", size=data_size, n=0)
    assert ds1.path == ds.path
    st = os.stat(ds1.path)
    m_time = st.st_mtime_ns
    assert m_time == m_time_ref

    ds2 = from_source("dummy-source", "zeros", force=_force_false, size=data_size, n=0)
    assert ds2.path == ds.path
    st = os.stat(ds2.path)
    m_time = st.st_mtime_ns
    assert m_time == m_time_ref

    ds3 = from_source("dummy-source", "zeros", force=_force_true, size=data_size, n=0)
    assert ds3.path == ds.path
    st = os.stat(ds3.path)
    m_time = st.st_mtime_ns
    assert m_time != m_time_ref
    m_time_ref = m_time

    ds4 = from_source("dummy-source", "zeros", size=data_size, n=0)
    assert ds4.path == ds.path
    st = os.stat(ds4.path)
    m_time = st.st_mtime_ns
    assert m_time == m_time_ref


if __name__ == "__main__":
    from earthkit.data.testing import main

    main(__file__)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np
import pytest
import yaml

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils import ensure_iterable


def to_tuple(x):
    return tuple(ensure_iterable(x))


def grid_list(files=None):
    with open(earthkit_test_data_file(os.path.join("xr_engine", "xr_grid.yaml")), "r") as f:
        r = yaml.safe_load(f)

    files = [] if files is None else files
    for item in r:
        if not files or item["file"] in files:
            yield (item["file"], item["dims"], item["coords"], item["distinct_ll"])


@pytest.mark.cache
@pytest.mark.parametrize(
    "file,dims,coords,distinct_ll",
    # grid_list(files=["sh_t32.grib1"]),
    grid_list(),
)
def test_xr_engine_grid(file, dims, coords, distinct_ll):
    ds = from_source("url", earthkit_remote_test_data_file("test-data", "xr_engine", "grid", file))

    a = ds.to_xarray()
    assert a is not None

    for k, v in dims.items():
        assert a.dims[k] == v

    for k, v in coords.items():
        assert a.coords[k].dims == to_tuple(v[0])
        assert a.coords[k].shape == to_tuple(v[1])

    assert a.t.dims[-(len(dims)) :] == tuple(dims.keys())

    if "latitude" in a.coords and "longitude" in a.coords:
        if distinct_ll:
            lat, lon = ds[0].metadata(("distinctLatitudes", "distinctLongitudes"))
        else:
            ll = ds[0].to_latlon()
            lat = ll["lat"]
            lon = ll["lon"]

        assert np.allclose(a.latitude.values, lat)
        assert np.allclose(a.longitude.values, lon)


@pytest.mark.cache
@pytest.mark.parametrize(
    "add_geo_coords",
    [False, True],
)
def test_xr_engine_add_geo_coords(add_geo_coords):
    ds = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "grid", "reduced_gg_O32.grib1")
    )

    a = ds.to_xarray(add_geo_coords=add_geo_coords)

    assert list(a.sizes.keys())[-1] == "values"

    if add_geo_coords:
        assert "latitude" in a.coords
        assert "longitude" in a.coords
    else:
        assert "latitude" not in a.coords
        assert "longitude" not in a.coords


@pytest.mark.cache
def test_xr_engine_gridspec():
    ds = from_source(
        "url", earthkit_remote_test_data_file("test-data", "xr_engine", "grid", "reduced_gg_O32.grib1")
    )

    r = ds.to_xarray()

    gs = r["r"].earthkit.grid_spec
    assert gs["type"] == "reduced_gg"
    assert gs["grid"] == "O32"

    gs = r["t"].earthkit.grid_spec
    assert gs["type"] == "reduced_gg"
    assert gs["grid"] == "O32"

    gs = r.earthkit.grid_spec
    assert gs["type"] == "reduced_gg"
    assert gs["grid"] == "O32"



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import pytest
from earthkit.utils.array import _TORCH
from earthkit.utils.testing import NO_TORCH
from earthkit.utils.testing import check_array_type

from earthkit.data import from_source
from earthkit.data.testing import earthkit_remote_test_data_file


@pytest.mark.skipif(NO_TORCH, reason="No pytorch installed")
@pytest.mark.cache
def test_xr_engine_torch_core():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))

    ds = ds_ek.to_xarray(array_backend="torch")
    check_array_type(ds["t"].data, _TORCH)


@pytest.mark.skipif(NO_TORCH, reason="No pytorch installed")
@pytest.mark.cache
def test_xr_engine_torch_core_compat():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))

    ds = ds_ek.to_xarray(array_module="torch")
    check_array_type(ds["t"].data, _TORCH)


@pytest.mark.skipif(NO_TORCH, reason="No pytorch installed")
@pytest.mark.cache
def test_xr_engine_torch_dtype():
    ds_ek = from_source("url", earthkit_remote_test_data_file("test-data/xr_engine/level/pl.grib"))

    def _check_dtype(dtype, expected_dtype):
        ds = ds_ek.to_xarray(array_backend="torch", dtype=dtype)
        assert ds["t"].data.dtype == expected_dtype

    dtype = _TORCH.float32
    expected_dtype = _TORCH.float32
    _check_dtype(dtype, expected_dtype)

    dtype = "float32"
    expected_dtype = _TORCH.float32
    _check_dtype(dtype, expected_dtype)

    dtype = _TORCH.float64
    expected_dtype = _TORCH.float64
    _check_dtype(dtype, expected_dtype)

    dtype = "float64"
    expected_dtype = _TORCH.float64
    _check_dtype(dtype, expected_dtype)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import os

from earthkit.data.readers.grib.codes import GribCodesMessagePositionIndex
from earthkit.data.readers.grib.index import GribFieldListInFiles
from earthkit.data.utils.parts import Part

LOG = logging.getLogger(__name__)


class GribFieldListInOneFile(GribFieldListInFiles):
    VERSION = 1

    @property
    def availability_path(self):
        return os.path.join(self.path, ".availability.pickle")

    def __init__(self, path, parts=None, positions=None, **kwargs):
        assert isinstance(path, str), path

        self.path = path
        self._file_parts = parts
        self.__positions = positions
        super().__init__(**kwargs)

    @property
    def _positions(self):
        if self.__positions is None:
            self.__positions = GribCodesMessagePositionIndex(self.path, self._file_parts)
        return self.__positions

    def part(self, n):
        return Part(self.path, self._positions.offsets[n], self._positions.lengths[n])

    def number_of_parts(self):
        return len(self._positions)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import warnings
from abc import abstractmethod
from functools import cached_property

from earthkit.data.core.geography import Geography
from earthkit.data.core.metadata import Metadata
from earthkit.data.core.metadata import MetadataAccessor
from earthkit.data.core.metadata import MetadataCacheHandler
from earthkit.data.core.metadata import WrappedMetadata
from earthkit.data.indexing.database import GRIB_KEYS_NAMES
from earthkit.data.readers.grib.gridspec import make_gridspec
from earthkit.data.utils.bbox import BoundingBox
from earthkit.data.utils.dates import datetime_from_grib
from earthkit.data.utils.dates import to_timedelta

LOG = logging.getLogger(__name__)


def missing_is_none(x):
    return None if x == 2147483647 else x


class GribFieldGeography(Geography):
    def __init__(self, metadata):
        self.metadata = metadata
        self.check_rotated_support()

    @cached_property
    def spectral(self):
        return self.metadata._handle.get("gridType", "") == "sh"

    def latitudes(self, dtype=None):
        r"""Return the latitudes of the field.

        Returns
        -------
        ndarray
        """
        return self.metadata._handle.get_latitudes(dtype=dtype)

    def longitudes(self, dtype=None):
        r"""Return the longitudes of the field.

        Returns
        -------
        ndarray
        """
        return self.metadata._handle.get_longitudes(dtype=dtype)

    def distinct_latitudes(self, dtype=None):
        return self.metadata._handle.get("distinctLatitudes", dtype=dtype)

    def distinct_longitudes(self, dtype=None):
        return self.metadata._handle.get("distinctLongitudes", dtype=dtype)

    def x(self, dtype=None):
        r"""Return the x coordinates in the field's original CRS.

        Returns
        -------
        ndarray
        """
        grid_type = self.metadata.get("gridType", None)
        if grid_type in ["regular_ll", "reduced_gg", "regular_gg"]:
            return self.longitudes(dtype=dtype)

    def y(self, dtype=None):
        r"""Return the y coordinates in the field's original CRS.

        Returns
        -------
        ndarray
        """
        grid_type = self.metadata.get("gridType", None)
        if grid_type in ["regular_ll", "reduced_gg", "regular_gg"]:
            return self.latitudes(dtype=dtype)

    def shape(self):
        r"""Get the shape of the field.

        For structured grids the shape is a tuple in the form of (Nj, Ni) where:

        - ni: the number of gridpoints in i direction (longitude for a regular latitude-longitude grid)
        - nj: the number of gridpoints in j direction (latitude for a regular latitude-longitude grid)

        For other grid types the number of gridpoints is returned as ``(num,)``

        Returns
        -------
        tuple
        """
        Nj = missing_is_none(self.metadata.get("Nj", None))
        Ni = missing_is_none(self.metadata.get("Ni", None))
        if Ni is None or Nj is None:
            n = self.metadata.get("numberOfDataPoints", None)
            return (n,)  # shape must be a tuple
        return (Nj, Ni)

    def _unique_grid_id(self):
        return self.metadata.get("md5GridSection", None)

    def projection(self):
        r"""Return information about the projection.

        Returns
        -------
        :obj:`Projection`

        Examples
        --------
        >>> import earthkit.data
        >>> ds = earthkit.data.from_source("file", "docs/examples/test.grib")
        >>> ds.projection()
        <Projected CRS: +proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to ...>
        Name: unknown
        Axis Info [cartesian]:
        - E[east]: Easting (unknown)
        - N[north]: Northing (unknown)
        - h[up]: Ellipsoidal height (metre)
        Area of Use:
        - undefined
        Coordinate Operation:
        - name: unknown
        - method: Equidistant Cylindrical
        Datum: Unknown based on WGS 84 ellipsoid
        - Ellipsoid: WGS 84
        - Prime Meridian: Greenwich
        >>> ds.projection().to_proj_string()
        '+proj=eqc +ellps=WGS84 +a=6378137.0 +lon_0=0.0 +to_meter=111319.4907932736 +no_defs +type=crs'
        """
        from earthkit.data.utils.projections import Projection

        return Projection.from_proj_string(self.metadata.get("projTargetString", None))

    def bounding_box(self):
        r"""Return the bounding box of the field.

        Returns
        -------
        :obj:`BoundingBox <data.utils.bbox.BoundingBox>`
        """
        return BoundingBox(
            north=self.metadata.get("latitudeOfFirstGridPointInDegrees", None),
            south=self.metadata.get("latitudeOfLastGridPointInDegrees", None),
            west=self.metadata.get("longitudeOfFirstGridPointInDegrees", None),
            east=self.metadata.get("longitudeOfLastGridPointInDegrees", None),
        )

    def gridspec(self):
        return make_gridspec(self.metadata)

    def resolution(self):
        grid_type = self.metadata.get("gridType")

        if grid_type in ("reduced_gg", "reduced_rotated_gg"):
            return self.metadata.get("gridName")

        if grid_type in ("regular_ll", "rotated_ll"):
            x = self.metadata.get("DxInDegrees")
            y = self.metadata.get("DyInDegrees")
            x = round(x * 1_000_000) / 1_000_000
            y = round(y * 1_000_000) / 1_000_000
            return x if x == y else None

        if grid_type in ["lambert", "lambert_azimuthal_equal_area"]:
            x = self.metadata.get("DxInMetres")
            y = self.metadata.get("DyInMetres")
            if x == y:
                return str(x / 1000).replace(".", "p") + "km"
            else:
                return None

        # raise ValueError(f"Unknown gridType={grid_type}")

    def mars_grid(self):
        if len(self.shape()) == 2:
            return [
                self.metadata.get("iDirectionIncrementInDegrees"),
                self.metadata.get("jDirectionIncrementInDegrees"),
            ]

        return self.metadata.get("gridName")

    def mars_area(self):
        north = self.metadata.get("latitudeOfFirstGridPointInDegrees")
        south = self.metadata.get("latitudeOfLastGridPointInDegrees")
        west = self.metadata.get("longitudeOfFirstGridPointInDegrees")
        east = self.metadata.get("longitudeOfLastGridPointInDegrees")
        return [north, west, south, east]

    @property
    def rotation(self):
        return (
            self.metadata.get("latitudeOfSouthernPoleInDegrees"),
            self.metadata.get("longitudeOfSouthernPoleInDegrees"),
            self.metadata.get("angleOfRotationInDegrees"),
        )

    @cached_property
    def rotated(self):
        grid_type = self.metadata.get("gridType")
        return "rotated" in grid_type

    @cached_property
    def rotated_iterator(self):
        return self.metadata.get("iteratorDisableUnrotate") is not None

    def check_rotated_support(self):
        if self.rotated and self.metadata.get("gridType") == "reduced_rotated_gg":
            from earthkit.data.utils.message import ECC_FEATURES

            if not ECC_FEATURES.version >= (2, 35, 0):
                raise RuntimeError("gridType=rotated_reduced_gg requires ecCodes >= 2.35.0")

    def latitudes_unrotated(self, **kwargs):
        if not self.rotated:
            return self.latitudes(**kwargs)

        if not self.rotated_iterator:
            try:
                from earthkit.geo.rotate import unrotate
            except ImportError:
                raise ImportError(
                    "GribFieldGeography.latitudes_unrotated requires 'earthkit-geo' to be installed"
                )

            grid_type = self.metadata.get("gridType")
            warnings.warn(f"ecCodes does not support rotated iterator for {grid_type}")
            lat, lon = self.latitudes(**kwargs), self.longitudes(**kwargs)
            south_pole_lat, south_pole_lon, _ = self.rotation
            lat, lon = unrotate(lat, lon, south_pole_lat, south_pole_lon)
            return lat

        with self.metadata._handle._set_tmp("iteratorDisableUnrotate", 1, 0):
            return self.latitudes(**kwargs)

    def longitudes_unrotated(self, **kwargs):
        if not self.rotated:
            return self.longitudes(**kwargs)

        if not self.rotated_iterator:
            try:
                from earthkit.geo.rotate import unrotate
            except ImportError:
                raise ImportError(
                    "GribFieldGeography.longitudes_unrotated requires 'earthkit-geo' to be installed"
                )

            grid_type = self.metadata.get("gridType")
            warnings.warn(f"ecCodes does not support rotated iterator for {grid_type}")
            lat, lon = self.latitudes(**kwargs), self.longitudes(**kwargs)
            south_pole_lat, south_pole_lon, _ = self.rotation
            lat, lon = unrotate(lat, lon, south_pole_lat, south_pole_lon)
            return lon

        with self.metadata._handle._set_tmp("iteratorDisableUnrotate", 1, 0):
            return self.longitudes(**kwargs)


class GribMetadata(Metadata):
    """GRIB metadata.

    :obj:`GribMetadata` is an abstract class and should not be instantiated directly.
    There are two concrete implementations: :class:`GribFieldMetadata` and
    :class:`StandAloneGribMetadata`.
    """

    LS_KEYS = [
        "centre",
        "shortName",
        "typeOfLevel",
        "level",
        "dataDate",
        "dataTime",
        "stepRange",
        "dataType",
        "number",
        "gridType",
    ]

    DESCRIBE_KEYS = [
        "shortName",
        "typeOfLevel",
        "level",
        "date",
        "time",
        "step",
        "number",
        "paramId",
        "marsClass",
        "marsStream",
        "marsType",
        "experimentVersionNumber",
    ]

    INDEX_KEYS = list(GRIB_KEYS_NAMES)

    NAMESPACES = [
        "default",
        "ls",
        "geography",
        "mars",
        "parameter",
        "statistics",
        "time",
        "vertical",
    ]

    ACCESSORS = {
        "valid_datetime": ["valid_datetime", "valid_time"],
        "gridspec": ["grid_spec", "gridspec"],
        "base_datetime": ["base_datetime", "forecast_reference_time", "base_time"],
        "reference_datetime": "reference_datetime",
        "indexing_datetime": ["indexing_time", "indexing_datetime"],
        "step_timedelta": "step_timedelta",
        "param_level": "param_level",
    }

    __handle_type = None

    def __init__(self, cache=None, **kwargs):
        self._cache = MetadataCacheHandler.make(cache)

    @staticmethod
    def _handle_type():
        """Return the expected handle type. Implemented like this
        to avoid cyclic import
        """
        if GribMetadata.__handle_type is None:
            from earthkit.data.readers.grib.codes import GribCodesHandle

            GribMetadata.__handle_type = GribCodesHandle
        return GribMetadata.__handle_type

    @abstractmethod
    def _handle(self):
        pass

    def __len__(self):
        return sum(map(lambda i: 1, self.keys()))

    def __contains__(self, key):
        return self._handle.__contains__(key)

    def __iter__(self):
        return self.keys()

    def keys(self):
        return self._handle.keys()

    def items(self):
        return self._handle.items()

    @MetadataCacheHandler.cache_get
    @MetadataAccessor(ACCESSORS)
    def get(self, key, default=None, *, astype=None, raise_on_missing=False):
        def _key_name(key):
            if key == "param":
                key = "shortName"
            elif key == "_param_id":
                key = "paramId"
            return key

        _kwargs = {}
        if not raise_on_missing:
            _kwargs["default"] = default

        # allow using the "grib." prefix.
        if key.startswith("grib."):
            key = key[5:]

        key = _key_name(key)

        v = self._handle.get(key, ktype=astype, **_kwargs)

        # special case when  "shortName" is "~".
        if key == "shortName" and v == "~":
            v = self._handle.get("paramId", ktype=str, **_kwargs)
        return v

    def _copy_key(self, target_handle, key):
        v_ori = self._handle.get(key, default=None)
        v_new = target_handle.get(key, default=None)
        if v_ori is not None and v_new is not None and v_ori != v_new:
            target_handle.set_long(key, v_ori)

    def override(self, *args, headers_only_clone=True, **kwargs):
        r"""Create a new metadata object by cloning a new GRIB handle and setting the keys in it.

        Parameters
        ----------
        *args: tuple
            Positional arguments. When present must be a dict with the GRIB keys to set in
            the new GRIB handle.
        headers_only_clone: bool, optional
            If True, the new GRIB handle will be created with headers_only=True to reduce the
            data section. With this the GRIB handle size will be significantly smaller, but the
            data section becomes unusable. Default is True.
        **kwargs: dict, optional
            Other keyword arguments specifying the GRIB keys to set.

        Returns
        -------
        :class:`WrappedMetadata`
            The new metadata object. There is always a :class:`StandAloneGribMetadata` object
            created containing the new GRIB handle updated with the specified keys.
            It is then wrapped in a :class:`WrappedMetadata` object storing ``"bitsPerValue"``
            as an extra key.


        Notes
        -----
        - When ``"bitsPerValue"`` is a key to set it is not written to the new handle. Instead, it
          is stored as an extra key in the resulting :class:`WrappedMetadata` object.
        """
        d = dict(*args, **kwargs)

        new_value_size = None
        gridspec = d.pop("gridspec", None)
        if gridspec is not None:
            from earthkit.data.readers.grib.gridspec import GridSpecConverter

            edition = d.get("edition", self["edition"])
            md, new_value_size = GridSpecConverter.to_metadata(gridspec, edition=edition)
            d.update(md)

        handle = self._handle.clone(headers_only=headers_only_clone)

        extra = {}

        # For the steps below consider the followings:
        # - we cannot reliably determine whether the original handle is reduced or not
        # - "bitsPerValue" needs a special treatment, because it cannot be set without
        #   repacking the data.
        # - we want to carry "bitsPerValue" over to the clone if possible

        # When headers_only=True, "bitsPerValue" in the clone is unreliable. Since we need to
        # carry "bitsPerValue" over ideally we should copy it into the clone but we
        # cannot do it since we just trimmed down the data section, so a proper repacking
        # is not possible. As a solution, we will generate a WrappedMetadata object and
        # store the original "bitsPerValue" in the extra dict.
        # When headers_only=False, we do not know whether the original handle was trimmed down
        # or not. Therefore, instead of applying complicated logic we follow the same
        # approach as for headers_only=True.
        key = "bitsPerValue"
        if key in d:
            extra[key] = d.pop(key)
        else:
            # we get the value form the original metadata object and not from the handle since
            # the handle can already be trimmed down
            v = self.get(key, default=None)
            if v is not None and v > 0:
                extra[key] = v
            # as a fallback we try to get the value from the clone
            else:
                v_clone = handle.get(key, None)
                if v_clone is not None and v_clone > 0:
                    extra[key] = v_clone

        if d:
            single = {}
            multiple = {}
            for k, v in d.items():
                if isinstance(v, (int, float, str, bool)):
                    single[k] = v
                else:
                    multiple[k] = v

            try:
                # Try to set all metadata at once
                # This is needed when we set multiple keys that are interdependent
                handle.set_multiple(single)
            except Exception as e:
                LOG.error("Failed to set metadata at once: %s", e)
                # Try again, but one by one
                for k, v in single.items():
                    handle.set(k, v)

            for k, v in multiple.items():
                handle.set(k, v)

        # we need to set the values to the new size otherwise the clone generated
        # with headers_only=True will be inconsistent
        if new_value_size is not None and new_value_size > 0:
            import numpy as np

            vals = np.zeros(new_value_size)
            handle.set_values(vals)

        # ensure that the cache settings are the same
        r = StandAloneGribMetadata(
            handle,
            cache=MetadataCacheHandler.clone_empty(self._cache),
        )

        if extra:
            r = WrappedMetadata(r, extra=extra)

        return r

    def namespaces(self):
        return self.NAMESPACES

    def as_namespace(self, namespace=None):
        r"""Return all the keys/values from a namespace.

        Parameters
        ----------
        namespace: str, None
            The :xref:`eccodes_namespace`. earthkit-data also defines the "default" namespace,
            which contains all the GRIB keys that ecCodes can access without specifying a namespace.
            When `namespace` is None or an empty :obj:`str` all the available
            keys/values are returned.

        Returns
        -------
        dict
            All the keys/values from the `namespace`.

        """
        if not isinstance(namespace, str) and namespace is not None:
            raise TypeError("namespace must be a str or None")

        if namespace == "default" or namespace == "":
            namespace = None
        r = self._handle.as_namespace(namespace)

        # special case when  "param" is "~".
        if r is not None:
            for k in ("shortName", "param"):
                if k in r and r[k] == "~":
                    r[k] = self.get(k)
        return r

    @cached_property
    def geography(self):
        return GribFieldGeography(self)

    def datetime(self):
        return {
            "base_time": self.base_datetime(),
            "valid_time": self.valid_datetime(),
        }

    def base_datetime(self):
        return self._datetime("dataDate", "dataTime")

    def valid_datetime(self):
        return self._datetime("validityDate", "validityTime")

    def reference_datetime(self):
        return self._datetime("referenceDate", "referenceTime")

    def indexing_datetime(self):
        return self._datetime("indexingDate", "indexingTime")

    def step_timedelta(self):
        v = self.get("endStep", None)
        if v is None:
            v = self.get("step", None)
        return to_timedelta(v)

    def _datetime(self, date_key, time_key):
        date = self.get(date_key, None)
        if date is not None:
            time = self.get(time_key, None)
            if time is not None:
                return datetime_from_grib(date, time)
        return None

    def param_level(self):
        return f"{self.get('shortName')}{self.get('level', default='')}"

    def dump(self, namespace=all, **kwargs):
        r"""Generate dump with all the metadata keys belonging to ``namespace``.

        In a Jupyter notebook it is represented as a tabbed interface.

        Parameters
        ----------
        namespace: :obj:`str`, :obj:`list`, :obj:`tuple`, :obj:`None` or :obj:`all`
            The namespace to dump. Any :xref:`eccodes_namespace` can be used here.
            earthkit-data also defines the "default" namespace, which contains all the GRIB keys
            that ecCodes can access without specifying a namespace. The following `namespace` values
            have a special meaning:

            - :obj:`all`: all the available namespaces will be used (including "default").
            - None or empty str: the "default" namespace will be used.

        **kwargs: dict, optional
            Other keyword arguments used for testing only

        Returns
        -------
        NamespaceDump
            Dict-like object with one item per namespace. In a Jupyter notebook represented
            as a tabbed interface to browse the dump contents.

        Examples
        --------
        :ref:`/examples/grib_metadata.ipynb`

        """
        from earthkit.data.utils.summary import format_namespace_dump

        if namespace is all:
            namespace = self.namespaces()
        else:
            if isinstance(namespace, str):
                namespace = [namespace]

        r = []
        for ns in namespace:
            v = self.as_namespace(ns)
            if v:
                r.append(
                    {
                        "title": ns if ns else "default",
                        "data": v,
                        "tooltip": f"Keys in the ecCodes {ns} namespace",
                    }
                )

        return format_namespace_dump(r, selected="parameter", details=self.__class__.__name__, **kwargs)

    def ls_keys(self):
        return self.LS_KEYS

    def describe_keys(self):
        return self.DESCRIBE_KEYS

    def index_keys(self):
        return self.INDEX_KEYS

    def data_format(self):
        return "grib"

    @property
    def gridspec(self):
        return self.geography.gridspec()

    def _make_restricted(self, r):
        return RestrictedGribMetadata(self)


class GribFieldMetadata(GribMetadata):
    """Represent the metadata of a GRIB field.

    :obj:`GribFieldMetadata` is created internally by a :obj:`GribField`. It does not
    own the ecCodes GRIB handle but can access it through the :obj:`GribField`.
    Calling :meth:`metadata` without arguments on a :obj:`GribField` returns this object.
    """

    def __init__(self, field, **kwargs):
        self._field = field
        assert field is not None
        super().__init__(**kwargs)

    @property
    def _handle(self):
        return self._field.handle

    def _hide_internal_keys(self):
        r = self.override()
        return RestrictedGribMetadata(r)


class StandAloneGribMetadata(GribMetadata):
    """Represent standalone GRIB metadata owning an ecCodes GRIB handle.

    :class:`StandAloneGribMetadata` possesses its own ecCodes handle. Calling
    :meth:`override` on :obj:`GribMetadata` always returns a
    :class:`StandAloneGribMetadata` object.

    >>> ds = earthkit.data.from_source("file", "docs/examples/test4.grib")
    >>> md = ds[0].metadata()
    >>> md["shortName"]
    't'
    >>> md.get("shortName")
    't'
    >>> md.get("nonExistentKey")
    >>> md.get("nonExistentKey", 12)
    12

    Examples
    --------
    :ref:`/examples/grib_metadata_object.ipynb`

    """

    def __init__(self, handle, **kwargs):
        if not isinstance(handle, self._handle_type()):
            raise TypeError(f"GribMetadata: expected handle type {self._handle_type()}, got {type(handle)}")
        self.__handle = handle
        super().__init__(**kwargs)

    @property
    def _handle(self):
        return self.__handle

    def _hide_internal_keys(self):
        return RestrictedGribMetadata(self)

    def __getstate__(self) -> dict:
        ret = {}
        ret["_handle"] = self._handle.get_buffer()
        # we do not serialize the cache contents
        ret["_cache"] = MetadataCacheHandler.serialise(self._cache)
        return ret

    def __setstate__(self, state: dict):
        from earthkit.data.readers.grib.memory import GribMessageMemoryReader

        self._cache = MetadataCacheHandler.deserialise(state.pop("_cache"))
        self.__handle = next(GribMessageMemoryReader(state.pop("_handle"))).handle


class RestrictedGribMetadata(WrappedMetadata):
    """Hide internal keys and namespaces in GRIB metadata.

    Examples
    --------
    :ref:`/examples/grib_metadata_object.ipynb`
    """

    EKD_NAMESPACE = ["grib"]

    # ideally bitsPerValue should be here. However, it is treated as an
    # extra key and cannot be an internal key.
    INTERNAL_KEYS = [
        "minimum",
        "maximum",
        "average",
        "standardDeviation",
        "skewness",
        "kurtosis",
        "min",
        "max",
        "avg",
        "sd",
        "skew",
        "kurt",
        "const",
        "isConstant",
        "numberOfMissing",
        "numberOfCodedValues",
        "bitmapPresent",
        "offsetValuesBy",
        "packingError",
        "referenceValue",
        "referenceValueError",
        "unpackedError",
        "values",
        # "numberOfPoints",
        # "numberOfDataPoints",
        "latLonValues",
    ]
    INTERNAL_NAMESPACES = ["statistics"]

    def __init__(self, metadata):
        super().__init__(
            metadata,
            hidden=self.INTERNAL_KEYS,
            hidden_namespaces=self.INTERNAL_NAMESPACES,
            enforced_namespaces=self.EKD_NAMESPACE,
        )

    def _hide_internal_keys(self):
        return self

    def __getstate__(self) -> dict:
        state = super().__getstate__()
        return state

    def __setstate__(self, state: dict):
        super().__setstate__(state)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import numpy as np
import pytest
from earthkit.utils.testing import check_array_type

import earthkit.data
from earthkit.data.testing import NO_GEO
from earthkit.data.testing import check_array
from earthkit.data.testing import earthkit_examples_file
from earthkit.data.testing import earthkit_test_data_file
from earthkit.data.utils import projections

here = os.path.dirname(__file__)
sys.path.insert(0, here)
from grib_fixtures import FL_NUMPY  # noqa: E402
from grib_fixtures import FL_TYPES  # noqa: E402
from grib_fixtures import load_grib_data  # noqa: E402


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize("index", [0, None])
def test_grib_to_latlon_single(fl_type, index):
    f, array_backend = load_grib_data("test_single.grib", fl_type, folder="data")

    eps = 1e-5
    g = f[index] if index is not None else f
    v = g.to_latlon(flatten=True)
    assert isinstance(v, dict)
    check_array_type(v["lon"], array_backend, dtype="float64")
    check_array_type(v["lat"], array_backend, dtype="float64")
    check_array(
        array_backend.to_numpy(v["lon"]),
        (84,),
        first=0.0,
        last=330.0,
        meanv=165.0,
        eps=eps,
    )
    check_array(
        array_backend.to_numpy(v["lat"]),
        (84,),
        first=90,
        last=-90,
        meanv=0,
        eps=eps,
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize("index", [0, None])
def test_grib_to_latlon_single_shape(fl_type, index):
    f, array_backend = load_grib_data("test_single.grib", fl_type, folder="data")

    g = f[index] if index is not None else f
    v = g.to_latlon()
    assert isinstance(v, dict)
    check_array_type(v["lon"], array_backend, dtype="float64")
    check_array_type(v["lat"], array_backend, dtype="float64")

    # x
    assert v["lon"].shape == (7, 12)
    for x in v["lon"]:
        assert np.allclose(array_backend.to_numpy(x), np.linspace(0, 330, 12))

    # y
    assert v["lat"].shape == (7, 12)
    for i, y in enumerate(v["lat"]):
        assert np.allclose(array_backend.to_numpy(y), np.ones(12) * (90 - i * 30))


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_grib_to_latlon_multi(fl_type, dtype):
    f, _ = load_grib_data("test.grib", fl_type)

    v_ref = f[0].to_latlon(flatten=True, dtype=dtype)
    v = f.to_latlon(flatten=True, dtype=dtype)
    assert isinstance(v, dict)
    assert v.keys() == v_ref.keys()

    assert isinstance(v, dict)
    assert np.allclose(v["lat"], v_ref["lat"])
    assert np.allclose(v["lon"], v_ref["lon"])
    assert v["lat"].dtype == dtype
    assert v["lon"].dtype == dtype


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_latlon_multi_non_shared_grid(fl_type):
    f1, _ = load_grib_data("test.grib", fl_type)
    f2, _ = load_grib_data("test4.grib", fl_type)
    f = f1 + f2

    with pytest.raises(ValueError):
        f.to_latlon()


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize("index", [0, None])
def test_grib_to_points_single(fl_type, index):
    f, array_backend = load_grib_data("test_single.grib", fl_type, folder="data")

    eps = 1e-5
    g = f[index] if index is not None else f
    v = g.to_points(flatten=True)
    assert isinstance(v, dict)
    check_array_type(v["x"], array_backend, dtype="float64")
    check_array_type(v["y"], array_backend, dtype="float64")
    check_array(
        array_backend.to_numpy(v["x"]),
        (84,),
        first=0.0,
        last=330.0,
        meanv=165.0,
        eps=eps,
    )
    check_array(
        array_backend.to_numpy(v["y"]),
        (84,),
        first=90,
        last=-90,
        meanv=0,
        eps=eps,
    )


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_points_unsupported_grid(fl_type):
    f, _ = load_grib_data("mercator.grib", fl_type, folder="data")
    with pytest.raises(ValueError):
        f[0].to_points()


@pytest.mark.parametrize("fl_type", FL_NUMPY)
@pytest.mark.parametrize("dtype", [np.float32, np.float64])
def test_grib_to_points_multi(fl_type, dtype):
    f, _ = load_grib_data("test.grib", fl_type)

    v_ref = f[0].to_points(flatten=True, dtype=dtype)
    v = f.to_points(flatten=True, dtype=dtype)
    assert isinstance(v, dict)
    assert v.keys() == v_ref.keys()

    assert isinstance(v, dict)
    assert np.allclose(v["x"], v_ref["x"])
    assert np.allclose(v["y"], v_ref["y"])
    assert v["x"].dtype == dtype
    assert v["y"].dtype == dtype


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_to_points_multi_non_shared_grid(fl_type):
    f1, _ = load_grib_data("test.grib", fl_type)
    f2, _ = load_grib_data("test4.grib", fl_type)
    f = f1 + f2

    with pytest.raises(ValueError):
        f.to_points()


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_bbox(fl_type):
    ds, _ = load_grib_data("test.grib", fl_type)
    bb = ds.bounding_box()
    assert len(bb) == 2
    for b in bb:
        assert b.as_tuple() == (73, -27, 33, 45)


@pytest.mark.parametrize("fl_type", FL_TYPES)
@pytest.mark.parametrize("index", [0, None])
def test_grib_projection_ll(fl_type, index):
    f, _ = load_grib_data("test.grib", fl_type)

    if index is not None:
        g = f[index]
    else:
        g = f
    assert isinstance(g.projection(), (projections.EquidistantCylindrical, projections.LongLat))


@pytest.mark.parametrize("fl_type", FL_TYPES)
def test_grib_projection_mercator(fl_type):
    f, _ = load_grib_data("mercator.grib", fl_type, folder="data")
    projection = f[0].projection()
    assert isinstance(projection, projections.Mercator)
    assert projection.parameters == {
        "true_scale_latitude": 20,
        "central_latitude": 0,
        "central_longitude": 0,
        "false_easting": 0,
        "false_northing": 0,
    }
    assert projection.globe == dict()


@pytest.mark.parametrize(
    "path,expected_value",
    [
        (earthkit_examples_file("test.grib"), 4.0),
        (earthkit_test_data_file("rgg_small_subarea_cellarea_ref.grib"), "O1280"),
        (earthkit_test_data_file("rotated_N32_subarea.grib"), "N32"),
        (earthkit_test_data_file("rotated_wind_20x20.grib"), 20),
        (earthkit_test_data_file("mercator.grib"), None),
        (earthkit_test_data_file("ll_10_20.grib"), None),
    ],
)
def test_grib_resolution(path, expected_value):
    ds = earthkit.data.from_source("file", path)

    if isinstance(expected_value, str):
        assert ds[0].resolution == expected_value
    elif expected_value is None:
        assert ds[0].resolution is None
    else:
        assert np.isclose(ds[0].resolution, expected_value)


@pytest.mark.parametrize(
    "path,expected_value",
    [
        (earthkit_examples_file("test.grib"), [73.0, -27.0, 33.0, 45.0]),
        (
            earthkit_test_data_file("rgg_small_subarea_cellarea_ref.grib"),
            [89.877, 36.233, 84.815, 46.185],
        ),
        (
            earthkit_test_data_file("rotated_N32_subarea.grib"),
            [26.511, 0.0, -12.558, 39.375],
        ),
        (
            earthkit_test_data_file("rotated_wind_20x20.grib"),
            [80.0, 0.0, -80.0, 340.0],
        ),
        (
            earthkit_test_data_file("mercator.grib"),
            [16.9775, 291.9722, 19.5221, 296.0156],
        ),
    ],
)
def test_grib_mars_area(path, expected_value):
    ds = earthkit.data.from_source("file", path)

    assert np.allclose(np.asarray(ds[0].mars_area), np.asarray(expected_value))


@pytest.mark.parametrize(
    "path,expected_value",
    [
        (earthkit_examples_file("test.grib"), [4.0, 4.0]),
        (
            earthkit_test_data_file("rgg_small_subarea_cellarea_ref.grib"),
            "O1280",
        ),
        (
            earthkit_test_data_file("rotated_N32_subarea.grib"),
            "N32",
        ),
        (
            earthkit_test_data_file("rotated_wind_20x20.grib"),
            [20.0, 20.0],
        ),
        (
            earthkit_test_data_file("mercator.grib"),
            [None, None],
        ),
    ],
)
def test_grib_mars_grid(path, expected_value):
    ds = earthkit.data.from_source("file", path)

    if isinstance(expected_value, str):
        assert ds[0].mars_grid == expected_value
    elif expected_value == [None, None]:
        assert ds[0].mars_grid == expected_value
    else:
        assert np.allclose(np.asarray(ds[0].mars_grid), np.asarray(expected_value))


@pytest.mark.skipif(NO_GEO, reason="No earthkit-geo support")
def test_grib_grid_points_rotated_ll():
    """The"""
    ds = earthkit.data.from_source("file", earthkit_test_data_file("rotated_wind_20x20.grib"))

    # grid points
    res = ds[0].grid_points()
    ref1 = np.array([30.0, 29.351052, 27.504876, 24.734374]), np.array(
        [140.0, 136.09296, 132.770576, 130.469424]
    )

    ref2 = np.array([-17.968188, -14.787578, -12.22927, -10.573044]), np.array(
        [-50.356844, -48.94784, -46.558096, -43.46374]
    )

    assert np.allclose(res[0][:4], ref1[0])
    assert np.allclose(res[1][:4], ref1[1])
    assert np.allclose(res[0][-4:], ref2[0])
    assert np.allclose(res[1][-4:], ref2[1])

    # unrotated grid points
    ds1 = earthkit.data.from_source("file", earthkit_test_data_file("wind_20x20.grib"))

    res = ds[0].grid_points_unrotated()
    ref = ds1[0].grid_points()

    assert np.allclose(res[0], ref[0])
    assert np.allclose(res[1], ref[1])


@pytest.mark.skipif(NO_GEO, reason="No earthkit-geo support")
def test_grib_grid_points_rotated_rgg():
    ds = earthkit.data.from_source("file", earthkit_test_data_file("rotated_N32_subarea.grib"))

    # grid points
    res = ds[0].grid_points()

    # front
    ref1 = np.array([85.489232, 84.81188, 83.171928, 81.086144]), np.array(
        [140.0, 110.950144, 92.460416, 82.07156]
    )

    # back
    ref2 = np.array([44.011184, 42.14694, 40.199948, 38.1796]), np.array(
        [4.244462, 7.003924, 9.575494, 11.973933]
    )

    assert np.allclose(res[0][:4], ref1[0])
    assert np.allclose(res[1][:4], ref1[1])
    assert np.allclose(res[0][-4:], ref2[0])
    assert np.allclose(res[1][-4:], ref2[1])

    # unrotated grid points
    res = ds[0].grid_points_unrotated()

    # front
    ref1 = np.array([26.510768, 26.51076943, 26.5107701, 26.51076846]), np.array(
        [1.28492181e-15, 2.81250046e00, 5.62500163e00, 8.43749805e00]
    )

    # back
    ref2 = np.array([-12.55775535, -12.55775697, -12.55775699, -12.5577565]), np.array(
        [30.93749931, 33.75000128, 36.56250084, 39.37500023]
    )

    assert np.allclose(res[0][:4], ref1[0])
    assert np.allclose(res[1][:4], ref1[1])
    assert np.allclose(res[0][-4:], ref2[0])
    assert np.allclose(res[1][-4:], ref2[1])


if __name__ == "__main__":
    from earthkit.data.testing import main

    main()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
from collections import defaultdict

from earthkit.data.core.fieldlist import FieldList
from earthkit.data.core.index import Selection
from earthkit.data.core.index import normalize_selection
from earthkit.data.core.order import build_remapping

LOG = logging.getLogger(__name__)


class CollectorJoiner:
    def __init__(self, func):
        self.func = func

    def format_name(self, x, **kwargs):
        return self.func(x, **kwargs)

    def format_string(self, x):
        return str(x)

    def join(self, args):
        remapped = "".join(str(x) for x in args)
        components = tuple([str(x) for x in args[1::2]])
        return (remapped, components)

    @staticmethod
    def patch(patch, value):
        if isinstance(value, tuple) and len(value) == 2:
            return patch(value[0]), value[1]
        return patch(value)


class IndexSelection(Selection):
    def match_element(self, element):
        return all(v(element) for k, v in self.actions.items())


class IndexDB:
    def __init__(self, index, component):
        # print(f"IndexDB: {index=}, {component=}")
        self._index = index if index is not None else dict()
        self._component = component if component is not None else dict()

    def index(self, key, maker=None):
        # LOG.debug(f"index(): {key=} {self._index=}")
        if key not in self._index:
            # # LOG.debug(f"Key={key} not found in IndexDB")
            if maker is not None:
                self._index[key] = maker(key)[0][key]
            else:
                raise KeyError(f"Could not find index for {key=}")
        return self._index[key]

    def component(self, key):
        if key not in self._component:
            raise KeyError(f"Could not find component for {key=}")
        return self._component[key]

    def collect(self, keys, component=False):
        remaining_keys = list(keys)
        indices = dict()
        components = dict()
        for k in keys:
            if k in self._index:
                indices[k] = self._index[k]
                remaining_keys.remove(k)
                if component and k in self._component:
                    components[k] = self._component[k]
        return remaining_keys, indices, components

    def filter(self, *args, **kwargs):
        kwargs = normalize_selection(*args, **kwargs)

        index = dict()
        component = dict()

        for k in self._index:
            if k in kwargs:
                selection = IndexSelection(dict(k=kwargs[k]))
                idx = list(i for i, element in enumerate(self._index[k]) if selection.match_element(element))
                index[k] = [self._index[k][i] for i in idx]
                if k in self._component:
                    component[k] = component[k][0], [self._component[k][1][i] for i in idx]
            else:
                index[k] = self._index[k]
                if k in self._component:
                    component[k] = self._component[k]
        return IndexDB(index, component)

    def __repr__(self) -> str:
        return f"IndexDB(_index={self._index}, component={self._component})"


class XArrayInputFieldList(FieldList):
    """
    A wrapper around a fieldlist that stores unique values.

    Only for internal use for building Xarray datasets.
    """

    def __init__(self, fieldlist, keys=None, db=None, remapping=None, scan_only=False, component=True):
        super().__init__()
        self.ds = fieldlist

        self.remapping = remapping
        if self.remapping is not None:
            self.remapping = build_remapping(remapping)

        self.db = IndexDB(None, None)
        if db is not None:
            self.db = db
        elif keys:
            self.db = IndexDB(*self.unique_values(keys, component=component))

        assert self.db

    def index(self, key, component=False):
        # print(f"called {key=}")
        if component:
            if self.remapping and key in self.remapping:
                return self.db.component(key)
            else:
                return None

        return self.db.index(key, self.unique_values)

    def __getitem__(self, n):
        return self.ds[n]

    def __len__(self):
        return len(self.ds)

    def make_releasable(self):
        self.ds = FieldList.from_fields([ReleasableField(f) for f in self.ds])

    def group(self, key, values):
        groups = defaultdict(list)
        for f in self.ds:
            v = str(f.metadata(key, remapping=self.remapping, default=None))
            if v in values:
                groups[v].append(f)

        for k, v in groups.items():
            if not v:
                continue
            db = None
            groups[k] = XArrayInputFieldList(FieldList.from_fields(v), db=db, remapping=self.remapping)

        return groups

    def sel(self, *args, **kwargs):
        assert "remapping" not in kwargs
        assert "patches" not in kwargs
        ds = self.ds.sel(*args, remapping=self.remapping, **kwargs)
        db = None
        # if not args and self.db and all(k in self.db for k in kwargs):
        #     db = self.db.filter(**kwargs)
        return XArrayInputFieldList(ds, db=db, remapping=self.remapping)

    def order_by(self, *args, **kwargs):
        if isinstance(self.ds, XArrayInputFieldList):
            kwargs.pop("remapping", None)

        assert "remapping" not in kwargs
        assert "patches" not in kwargs

        if isinstance(self.ds, XArrayInputFieldList):
            ds = self.ds.order_by(*args, **kwargs)
            return ds
        else:
            ds = self.ds.order_by(*args, remapping=self.remapping, **kwargs)
            ds = XArrayInputFieldList(
                ds,
                db=self.db,
                remapping=self.remapping,
            )
            return ds

    def unique_values(self, names, component=False):
        if isinstance(names, str):
            names = [names]

        keys, indices, components = self.db.collect(names, component=component)

        if keys:
            vals = defaultdict(dict)
            if component:
                components = dict()
                joiner = CollectorJoiner
            else:
                joiner = None

            for f in self.ds:
                r = f._attributes(keys, remapping=self.remapping, joiner=joiner)
                for k, v in r.items():
                    vals[k][v] = True

            vals = {k: tuple(values.keys()) for k, values in vals.items()}

            for k, v in vals.items():
                v = [x for x in v if x is not None]
                if all(isinstance(x, (int, datetime.timedelta)) for x in v):
                    vals[k] = sorted(v)
                else:
                    vals[k] = sorted(v, key=str)

            if component and self.remapping:
                for k, v in vals.items():
                    if k in self.remapping:
                        indices[k] = [x[0] for x in v]
                        components[k] = self.remapping.components(k), [x[1] for x in v]
                        assert len(indices[k]) == len(
                            components[k][1]
                        ), f"{len(indices[k])} != {len(components[k])} {indices[k]=} {components[k]=}"
                    else:
                        indices[k] = v
            else:
                for k, v in vals.items():
                    indices[k] = v

        if component:
            return indices, components
        else:
            return indices, None

    def unwrap(self):
        ds = self.ds
        while isinstance(ds, XArrayInputFieldList):
            ds = ds.ds
        return ds

    def __getstate__(self):
        """As a simplification, only serialise the unwrapped fieldlist.
        We can assume that when there is a need for serialisation the wrapper
        structure can be discarded.
        """
        r = {}
        r["ds"] = self.unwrap()
        return r

    def __setstate__(self, state):
        self.ds = state["ds"]
        self.db = None
        self.remapping = None


class ReleasableField:
    def __init__(self, field):
        self.field = field
        self.keep = True
        self.released = False

    def to_numpy(self, *args, **kwargs):
        if self.released:
            return None

        r = self.field.to_numpy(*args, **kwargs)
        if not self.keep:
            self._release()
        return r

    def to_array(self, *args, **kwargs):
        if self.released:
            return None

        r = self.field.to_array(*args, **kwargs)
        if not self.keep:
            self._release()
        return r

    def _release(self):
        if not self.released:
            if hasattr(self.field, "_release"):
                self.field._release()
            self.released = True

    def metadata(self, *args, **kwargs):
        if self.released:
            return None
        return self.field.metadata(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.field, name)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging

from multiurl import Downloader

from earthkit.data.core.caching import cache_file
from earthkit.data.core.config import CONFIG
from earthkit.data.core.statistics import record_statistics
from earthkit.data.utils.parts import PathAndParts
from earthkit.data.utils.progbar import progress_bar

from .file import FileSource

LOG = logging.getLogger(__name__)


def _ignore(*args, **kwargs):
    pass


SCHEME_TO_SOURCE = {"ec:": "ecfs", "ecfs:": "ecfs"}


def download_and_cache(
    url,
    *,
    owner="url",
    parts=None,
    verify=True,
    force=None,
    chunk_size=1024 * 1024,
    range_method="auto",
    http_headers=None,
    update_if_out_of_date=False,
    fake_headers=None,  # When HEAD is not allowed but you know the size
    **kwargs,
):
    # TODO: re-enable this feature
    extension = None

    LOG.debug("URL %s", url)

    downloader = Downloader(
        url,
        chunk_size=chunk_size,
        timeout=CONFIG.get("url-download-timeout"),
        verify=verify,
        parts=parts,
        range_method=range_method,
        http_headers=http_headers,
        fake_headers=fake_headers,
        statistics_gatherer=record_statistics,
        progress_bar=progress_bar,
        resume_transfers=True,
        override_target_file=False,
        download_file_extension=".download",
    )

    if extension and extension[0] != ".":
        extension = "." + extension

    if extension is None:
        extension = downloader.extension()

    path = downloader.local_path()
    if path is not None:
        return

    def out_of_date(url, path, cache_data):
        if CONFIG.get("check-out-of-date-urls") is False:
            return False

        if downloader.out_of_date(path, cache_data):
            if CONFIG.get("download-out-of-date-urls") or update_if_out_of_date:
                LOG.warning(
                    "Invalidating cache version and re-downloading %s",
                    url,
                )
                return True
            else:
                LOG.warning(
                    "To enable automatic downloading of updated URLs set the 'download-out-of-date-urls'"
                    " setting to True",
                )
        return False

    if force is None:
        force = out_of_date

    def download(target, _):
        downloader.download(target)
        return downloader.cache_data()

    path = cache_file(
        owner,
        download,
        dict(url=url, parts=parts),
        extension=extension,
        force=force,
    )

    return path


class UrlSourcePathAndParts(PathAndParts):
    compress = True


class UrlBase(FileSource):
    def __init__(
        self,
        url,
        chunk_size=1024 * 1024,
        parts=None,
        filter=None,
        merger=None,
        verify=True,
        range_method="auto",
        http_headers=None,
        fake_headers=None,  # When HEAD is not allowed but you know the size
        stream=False,
        auth=None,
        **kwargs,
    ):
        super().__init__(filter=filter, merger=merger)

        from earthkit.data.utils.url import UrlSpec
        from earthkit.data.utils.url import UrlSpecItem

        if isinstance(url, UrlSpecItem):
            self.url_spec = UrlSpec.from_urls(url)
        else:
            url_kwargs = dict(
                chunk_size=chunk_size,
                parts=parts,
                verify=verify,
                range_method=range_method,
                http_headers=http_headers,
                fake_headers=fake_headers,
                auth=auth,
            )
            self.url_spec = UrlSpec.from_urls(url, **url_kwargs)

        self._kwargs = kwargs
        LOG.debug(f"url={self.url} url_parts={self.url_parts} auth={auth} _kwargs={self._kwargs}")

    def connect_to_mirror(self, mirror):
        return mirror.connection_for_url(self, self.url, self.url_parts)

    def prepare_headers(self, url):
        headers = {}
        if self.http_headers is not None:
            headers = dict(self.http_headers)

        if self.auth is not None:
            headers.update(self.auth.auth_header(url))

        if not headers:
            headers = None

        return headers

    @property
    def url(self):
        return self.url_spec.url

    @property
    def url_parts(self):
        return self.url_spec.parts

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.url})"


class Url(UrlBase):
    """Represent a URL source.

    Parameters
    ----------
    url: str, list, tuple, UrlSpecItem, dict or an iterable of these
        A single ``url`` can be:

        * a str
        * a list/tuple of two items. The first item is the url as a str, while the second
          item defines the parts for the given url. In this case the ``parts`` kwargs
          cannot be used.
        * a UrlSpecItem
        * a dict with same content as UrlSpecItem. Mandatory key is "url" and optionally
          "parts" and "kwargs" can be used.

        Multiple urls can be provided as a list/tuple of of the single url formats above.
        Mixing single url formats in a list/tuple is not allowed.
    """

    def __init__(
        self,
        url,
        *,
        update_if_out_of_date=False,
        force=None,
        stream=False,
        **kwargs,
    ):
        super().__init__(url, **kwargs)

        self.other_source = None

        if isinstance(url, str):
            for scheme, source in SCHEME_TO_SOURCE.items():
                if url.startswith(scheme):
                    self.other_source = (source, url, kwargs)
                    return

        self.update_if_out_of_date = update_if_out_of_date
        self.force = force
        self.stream = stream

        if not self.stream:
            self._download()

    def mutate(self):

        if self.other_source:
            from earthkit.data import from_source

            source, url, kwargs = self.other_source
            return from_source(source, url, **kwargs)

        if self.stream:
            s = []
            # create one stream source per url
            for x in self.url_spec:
                s.append(SingleUrlStream(x))

            from .stream import make_stream_source_from_other

            return make_stream_source_from_other(s, **self._kwargs)

        return super().mutate()

    def _download(self):
        # TODO: re-enable this feature
        extension = None

        LOG.debug(
            (
                f"url={self.url} url_parts={self.url_parts} "
                # f"http_headers={self.http_headers}"
                f" _kwargs={self._kwargs}"
            )
        )

        self.downloader = Downloader(
            self.url_spec.zipped(),
            timeout=CONFIG.get("url-download-timeout"),
            statistics_gatherer=record_statistics,
            progress_bar=progress_bar,
            resume_transfers=True,
            override_target_file=False,
            download_file_extension=".download",
            **self.url_spec[0].kwargs,
        )

        if extension and extension[0] != ".":
            extension = "." + extension

        if extension is None:
            extension = self.downloader.extension()

        self.path = self.downloader.local_path()
        if self.path is not None:
            return

        if self.force is None:
            self.force = self.out_of_date

        def download(target, _):
            self.downloader.download(target)
            return self.downloader.cache_data()

        self.path = self.cache_file(
            download,
            dict(url=self.url, parts=self.url_parts),
            extension=extension,
            force=self.force,
        )

        # cache data may contain the result of the http HEAD request
        h = self.downloader.cache_data()
        if isinstance(h, dict):
            self.content_type = h.get("content-type")

    def out_of_date(self, url, path, cache_data):
        if CONFIG.get("check-out-of-date-urls") is False:
            return False

        if self.downloader.out_of_date(path, cache_data):
            if CONFIG.get("download-out-of-date-urls") or self.update_if_out_of_date:
                LOG.warning(
                    "Invalidating cache version and re-downloading %s",
                    self.url,
                )
                return True
            else:
                LOG.warning(
                    "To enable automatic downloading of updated URLs set the 'download-out-of-date-urls'"
                    " setting to True",
                )
        return False


class SingleUrlStream(UrlBase):
    def __init__(
        self,
        url,
        **kwargs,
    ):
        super().__init__(url, **kwargs)

        if len(self.url_spec) != 1:
            raise TypeError("Only a single url is supported")

        assert isinstance(self.url[0], str)

        from urllib.parse import urlparse

        assert isinstance(self.url, (list, tuple)), f"{self.url=}"
        o = urlparse(self.url[0])
        if o.scheme not in ("http", "https"):
            raise NotImplementedError(f"Streams are not supported for scheme={o.scheme} urls")

    def mutate(self):
        from .stream import _from_source

        return _from_source(self, **self._kwargs)

    def to_stream(self):
        from earthkit.data.utils.stream import RequestIterStreamer

        downloader = Downloader(
            self.url_spec.zipped(),
            timeout=CONFIG.get("url-download-timeout"),
            # verify=self.verify,
            # range_method=self.range_method,
            # http_headers=self.prepare_headers(self.url),
            # fake_headers=self.fake_headers,
            statistics_gatherer=_ignore,
            progress_bar=progress_bar,
            resume_transfers=False,
            override_target_file=False,
            **self.url_spec[0].kwargs,
        )

        _, _, _, _ = downloader.estimate_size(None)

        # cache data may contain the result of the http HEAD request
        h = downloader.cache_data()
        if isinstance(h, dict):
            self.content_type = h.get("content-type")

        stream = downloader.make_stream()
        return RequestIterStreamer(stream(chunk_size=self.url_spec[0].kwargs["chunk_size"]))


source = Url



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from . import EncodedData
from . import Encoder

LOG = logging.getLogger(__name__)


class NetCDFEncodedData(EncodedData):
    # Prefer file path when writing to netCDF files. The reason is as follows:
    #
    # The Xarray.to_netcdf() documentation states that:
    # ... file-like objects are only supported by the scipy engine.
    # If no path is provided, this function returns the resulting netCDF file as bytes;
    # in this case, we need to use scipy, which does not support netCDF version 4
    # (the default format becomes NETCDF3_64BIT).
    prefer_file_path = True

    def __init__(self, ds):
        self.ds = ds

    def to_bytes(self):
        return self.ds.to_netcdf(None)

    def to_file(self, file):
        if hasattr(self.ds, "earthkit"):
            self.ds.earthkit.to_netcdf(file)

        else:
            self.ds.to_netcdf(file)

    def metadata(self, key):
        raise NotImplementedError


class NetCDFEncoder(Encoder):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def encode(
        self,
        data=None,
        **kwargs,
    ):
        if data is not None:
            from earthkit.data.wrappers import get_wrapper

            data = get_wrapper(data, fieldlist=False)
            return data._encode(self, **kwargs)
        else:
            raise ValueError("No data to encode")

    def _encode(
        self,
        data=None,
        values=None,
        min=None,
        max=None,
        check_nans=False,
        metadata={},
        template=None,
        # return_bytes=False,
        missing_value=9999,
        **kwargs,
    ):
        _kwargs = kwargs.copy()
        if data is not None:
            # TODO: find better way to check if the earthkit engine is used
            if hasattr(data, "to_xarray_earthkit"):
                earthkit_to_xarray_kwargs = _kwargs.pop("earthkit_to_xarray_kwargs", {})
                earthkit_to_xarray_kwargs["add_earthkit_attrs"] = False
                _kwargs = earthkit_to_xarray_kwargs
        else:
            _kwargs = {}

        return NetCDFEncodedData(data.to_xarray(**_kwargs))

    def _encode_field(self, field, **kwargs):
        return self._encode(field, **kwargs)

    def _encode_fieldlist(self, data, **kwargs):
        return self._encode(data, **kwargs)

    def _encode_xarray(self, data, **kwargs):
        return NetCDFEncodedData(data)


encoder = NetCDFEncoder



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import subprocess

from earthkit.data.core.config import CONFIG
from earthkit.data.core.temporary import temp_file

from .ecmwf_api import ECMWFApi
from .ecmwf_api import MARSAPIKeyPrompt

LOG = logging.getLogger(__name__)


def _no_log(msg):
    pass


class StandaloneMarsClient:
    def __init__(self, log="default"):
        self.log = log

    def execute(self, request, target, log=None):
        req = ["retrieve,"]

        for k, v in request.items():
            if isinstance(v, (list, tuple)):
                v = "/".join([str(x) for x in v])
            req += [f"{k}={v},"]

        req += [f'target="{target}"']
        req_str = "\n".join(req)
        with temp_file() as filename:
            with open(filename, "w") as f:
                f.write(req_str + "\n")
            LOG.debug(f"Sending Mars request: '{req_str}'")

            log = {}
            if self.log is None:
                log = {"stdout": subprocess.DEVNULL, "stderr": subprocess.DEVNULL}
            elif self.log and isinstance(self.log, dict):
                log = self.log
            elif self.log != "default":
                raise ValueError(f"Unsupported log type={type(self.log)}")

            subprocess.run(
                [self.command(), filename],
                env=dict(os.environ, MARS_AUTO_SPLIT_BY_DATES="1"),
                check=True,
                **log,
            )

    @staticmethod
    def command():
        return os.environ.get("MARS_CLIENT_COMMAND", "/usr/local/bin/mars")

    @staticmethod
    def enabled():
        return CONFIG.get("use-standalone-mars-client-when-available") and os.path.exists(
            StandaloneMarsClient.command()
        )


class MarsRetriever(ECMWFApi):
    def service(self):
        if StandaloneMarsClient.enabled():
            return StandaloneMarsClient(self.log)

        kwargs = {}
        if self.log is None:
            kwargs = {"log": _no_log}
        elif callable(self.log):
            kwargs = {"log": self.log}
        elif self.log != "default":
            raise ValueError(f"Unsupported log type={type(self.log)}")

        try:
            import ecmwfapi
        except ImportError:
            raise ImportError("MARS access requires 'ecmwf-api-client' to be installed")

        if self.prompt:
            prompt = MARSAPIKeyPrompt()
            prompt.check()

            try:
                return ecmwfapi.ECMWFService("mars", **kwargs)
            except Exception as e:
                if ".ecmwfapirc" in str(e) or not self.prompt.has_config_env():
                    LOG.warning(e)
                    LOG.exception(f"Could not load ecmwf api (mars) client. {e}")
                    prompt.ask_user_and_save()
                else:
                    raise

        return ecmwfapi.ECMWFService("mars", **kwargs)


source = MarsRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def drop_unwanted_series(df, key=None, axis=1):
    if df is None:
        return

    # only show the column for number in the default set of keys if
    # there are any valid values in it
    r = None
    if axis == 1 and key in df.columns:
        r = df[key].unique()
    elif axis == 0 and key in df.index:
        r = df.loc[key].unique()
    if r is not None and len(r) == 1 and r[0] in ["0", None]:
        df.drop(key, axis=axis, inplace=True)


def format_list(v, full=False):
    if isinstance(v, list):
        if full is True:
            return ",".join([str(x) for x in v])
        else:
            if len(v) == 1:
                return v[0]
            if len(v) > 2:
                return ",".join([str(x) for x in [v[0], v[1], "..."]])
            else:
                return ",".join([str(x) for x in v])
    else:
        return v


def make_unique(x, full=False):
    v = set(x)
    r = [str(t) for t in v]
    return format_list(r, full=full)


def ls(metadata_proc, default_keys, n=None, keys=None, extra_keys=None, **kwargs):
    # do_print = kwargs.pop("print", False)

    if kwargs:
        raise TypeError(f"ls: unsupported arguments={kwargs}")

    _keys = {}
    if isinstance(default_keys, (list, tuple)):
        default_keys = {k: k for k in default_keys}

    _keys = dict(default_keys) if keys is None else keys
    if isinstance(_keys, (list, tuple)):
        _keys = {k: k for k in keys}
    elif isinstance(_keys, str):
        _keys = {keys: keys}

    if extra_keys is not None:
        if isinstance(extra_keys, (list, tuple)):
            if len(extra_keys) > 0:
                _keys.update({k: k for k in extra_keys})
        elif isinstance(extra_keys, str):
            _keys.update({extra_keys: extra_keys})
        elif isinstance(extra_keys, dict):
            _keys.update(extra_keys)

    if n == 0:
        raise ValueError("n cannot be 0")

    return format_ls(metadata_proc(_keys, n))


def format_ls(attributes):
    import pandas as pd

    df = pd.DataFrame.from_records(attributes)
    return df


def format_describe(attributes, *args, **kwargs):
    # TODO: this is GRIB specific code, should not be here
    import pandas as pd

    param = args[0] if len(args) == 1 else None
    if param is None:
        param = kwargs.pop("param", None)

    # do_print = kwargs.pop("print", True)

    df = pd.DataFrame.from_records(attributes, **kwargs)

    if df is None or df.empty:
        return None

    # these keys are changed in the output
    labels = {"marsClass": "class", "marsStream": "stream", "marsType": "type"}

    no_header = False
    main_axis = 1
    if param is None:
        df = df.groupby(["shortName", "typeOfLevel"]).agg(make_unique)
        df.rename(labels, axis=1, inplace=True)
    else:
        param_names = {int: "paramId", str: "shortName"}
        param_name = param_names.get(type(param), None)
        if param_name is not None:
            df = pd.DataFrame(df[df[param_name] == param].agg(make_unique, full=True).T)
            df.rename(labels, axis=0, inplace=True)
            no_header = True
            main_axis = 0
        else:
            return pd.DataFrame()

    drop_unwanted_series(df, key="number", axis=main_axis)

    df = df.style.set_properties(**{"text-align": "left"})
    df.set_table_styles([dict(selector="th", props=[("text-align", "left")])])
    if no_header:
        df.hide(axis="columns")

    return df


class NamespaceDump(dict):
    def __init__(self, data, **kwargs):
        d = {item["title"]: item["data"] for item in data}
        super().__init__(d)
        self.data = data
        self._kwargs = kwargs

    def _repr_html_(self):
        from earthkit.data.core.ipython import ipython_active

        if ipython_active:
            from earthkit.data.utils.html import tab
            from earthkit.data.utils.html import table_from_dict

            if len(self.data) == 1:
                return table_from_dict(self.data[0]["data"])
            elif len(self.data) > 1:
                for item in self.data:
                    item["text"] = table_from_dict(item["data"])

                return tab(self.data, **self._kwargs)
        else:
            return str(self)


def format_namespace_dump(data, selected=None, details=None, **kwargs):
    raw = kwargs.pop("_as_raw", False)

    if not raw:
        return NamespaceDump(data, selected=selected, details=details)
    else:
        return data


class BUFRTree:
    """Restructures the result of the bufr_dump ecCodes command."""

    def __init__(self, data, subset, compressed, uncompressed):
        self.data = data
        self.subset = subset
        self.compressed = compressed
        self.uncompressed = uncompressed

    def make_tree(self):
        """Restructures the the result of json bufr_dump into a format better
        suited for generating a tree view out of it.

        Returns
        -------
        dict
        """
        return self._load_dump()

    def _load_dump(self):
        r = {"header": [], "data": []}

        # data = data["messages"][0]

        for i, v in enumerate(self.data):
            # print(v)
            k = v["key"]
            val = v["value"]
            if isinstance(val, list):
                val = "array"

            r["header"].append({"key": k, "value": val})

            # start data section
            if k == "unexpandedDescriptors":
                # for uncompressed data tries to shown only the branch
                # with given subset
                if self.subset is not None and self.subset > 0 and self.uncompressed:
                    try:
                        subset_index = self.subset - 1
                        d = self.data[i + 1][subset_index][0]
                        if d["key"] == "subsetNumber" and d["value"] == self.subset:
                            self._parse_dump(self.data[i + 1][subset_index], r["data"])
                            break
                    except Exception:
                        pass
                self._parse_dump(self.data[i + 1], r["data"])
                break
        return r

    def _parse_dump(self, data, parent):
        arrayCnt = 0
        keyCnt = 0
        for v in data:
            if isinstance(v, list):
                arrayCnt += 1
            else:
                keyCnt += 1

        for v in data:
            if not isinstance(v, list):
                vals = v["value"]
                if isinstance(vals, list):
                    vals = self._format_list(vals)
                parent.append(
                    {
                        "key": v["key"],
                        "value": vals,
                        "units": v.get("units", None),
                    }
                )
            else:
                if arrayCnt > 1:
                    item = []
                    parent.append(item)
                    self._parse_dump(v, item)
                else:
                    self._parse_dump(v, parent)

    def _format_list(self, vals):
        if len(vals) > 0:
            if self.subset is not None and self.subset > 0:
                index = self.subset - 1
                if index < len(vals):
                    return f"{vals[index]} ({len(vals)} items)"
                else:
                    return f"[{vals[0]}, ...] ({len(vals)-1} items)"
            else:
                return f"[{vals[0]}, ...] ({len(vals)} items)"
        else:
            return "[]"


class BUFRHtmlTree:
    def make_html(self, data):
        """Generates a html/css tree view from the input representing the
        result of bufr_dump.

        Parameters
        ----------
        data: dict

        Returns
        -------
        str
        """
        return self._build(data)

    def _build(self, data):
        td1 = self._node(data["header"], [])
        td2 = self._node(data["data"], [])

        td = self._top_node("header", td1) + self._top_node("data", td2)
        t = f"""
        <ul class="tree">
        <li>
            {td}
        </li>
        </ul>
        """
        return t

    def _leaf(self, k, v, units):
        if units is not None:
            return f"""<li>{k}: {v} [{units}]</li>"""
        else:
            return f"""<li>{k}: {v}</li>"""

    def _top_node(self, name, v):
        return f"""
        <li>
        <details>
            <summary>{name}</summary>
            <ul>
            {v}
            </ul>
            </details>
        </li>
        """

    def _start_node(self, k, v):
        return f"""
        <li>
        <details>
            <summary>{k}: {v}</summary>
            <ul>
        """

    def _end_node(self):
        return """</ul>
            </details>
        </li>
        """

    def _node(self, data, parent):
        td = ""
        if isinstance(data, list):
            if parent:
                td += self._start_node(data[0]["key"], data[0]["value"])
            for v in data:
                td += self._node(v, data)
            if parent:
                td += self._end_node()
        else:
            td += self._leaf(data["key"], data["value"], data.get("units", None))

        return td


def make_bufr_html_tree(data, title, subset, compressed, uncompressed):
    from earthkit.data.core.ipython import ipython_active

    tree = BUFRTree(data, subset, compressed, uncompressed).make_tree()
    if ipython_active:
        from IPython.display import HTML

        from earthkit.data.utils.html import css

        t = ""
        if title is not None and title:
            t = f"""<div class="eh-description">{title}</div>"""

        t += BUFRHtmlTree().make_html(tree)
        style = css("tree")
        return HTML(style + t)

    return tree


def ncdump(path):
    import subprocess

    result = subprocess.run(["ncdump", "-h", path], stdout=subprocess.PIPE)
    t = result.stdout.decode("utf-8").split("\n")
    for line in t:
        print(line)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

LOG = logging.getLogger(__name__)


def _match_magic(magic, deeper_check):
    if magic is not None:
        type_id = b"BUFR"
        if not deeper_check:
            return len(magic) >= 4 and magic[:4] == type_id
        else:
            return type_id in magic
    return False


def reader(source, path, *, magic=None, deeper_check=False, **kwargs):
    if _match_magic(magic, deeper_check):
        from .bufr import BUFRReader

        parts = source.parts if hasattr(source, "parts") else None
        return BUFRReader(source, path, parts=parts)



import earthkit.data as ekd
from earthkit.data.readers.grib.output import new_grib_coder

ds = ekd.from_source("sample", "test.grib")

values = ds[1].values + 1

# case 1
c = new_grib_coder()
# c is a GribCodesHandle
d = c.encode(values, shortName="2t", template=ds[1])

# case 2
c = new_grib_coder(template=ds[1])
# c is a GribCodesHandle
d = c.encode(values, shortName="2t")
